import numpy as np
def find_dof_map(fes, idx1, idx2, trans = None):
   '''
   find_dof_map

   projection coupling between DoFs. so far, it is tested
   only for triangle Nedelec boundary elemets.

   it maps element nodes of on bdr1 (attribute = idx1) to 
   bdr2(attribute =idx2)

   boundary elements are projected to a 2D surface 
   and matched on the surface (using trans in keyword). 
   By default, it projects DoF points on y-z plan. 

   Need to test that this approach works even when two 
   surface is not parallel...

   '''
   if trans is None:
       trans = np.array([[0, 1, 0], [0, 0, 1]]) # projection to y, z         
   def get_vshape(fes, k1, idx):
       tr1 = fes.GetBdrElementTransformation(k1)
       el = fes.GetBE(k1)
       nodes1 = el.GetNodes()
       m = mfem.DenseMatrix(nodes1.GetNPoints(), tr1.GetSpaceDim())
       tr1.SetIntPoint(nodes1.IntPoint(idx))
       el.CalcVShape(tr1,m)
       return m.GetDataArray()[idx, :]
   def get_all_intpoint(fes, ibdr1, ibdr2, trans):
       pt = []
       subvdofs1 = []
       subvdofs2 = []
       map = scipy.sparse.lil_matrix((fes.GetNDofs(), fes.GetNDofs()), dtype=float)

       for k1, k2 in zip(ibdr1, ibdr2):
           tr1 = fes.GetBdrElementTransformation(k1)
           nodes1 = fes.GetBE(k1).GetNodes()
           vdof1 = fes.GetBdrElementVDofs(k1)
           pt1 = [np.dot(trans, tr1.Transform(nodes1.IntPoint(kk))) 
                 for kk in range(len(vdof1))]
           subvdof1 = [x if x>= 0 else -1-x for x in vdof1]
           vdof2 = fes.GetBdrElementVDofs(k2)
           nodes2 = fes.GetBE(k2).GetNodes()
           tr2 = fes.GetBdrElementTransformation(k2)
           pt2 = [np.dot(trans, tr2.Transform(nodes2.IntPoint(kk))) 
                 for kk in range(len(vdof2))]
           subvdof2 = [x if x>= 0 else -1-x for x in vdof2]

           newk1 = [(k, xx[0], xx[1]) for k, xx in enumerate(zip(vdof1, subvdof1)) 
                   if not xx[1] in subvdofs1]
           newk2 = [(k, xx[0], xx[1]) for k, xx in enumerate(zip(vdof2, subvdof2)) 
                   if not xx[1] in subvdofs2]
           pt1 = [pt1[kk] for kk, v, s in newk1]
           pt2 = [pt2[kk] for kk, v, s in newk2]

           #print subvdof1, newk1
           #print subvdof2, newk2
           for k, p in enumerate(pt1):
               dist = np.sum((pt2-p)**2, 1)
               d = np.where(dist == np.min(dist))[0]
               if len(d) == 1:
                   '''
                   this factor is not always 1
                   '''
                   s = np.sign(newk1[k][1] +0.5)*np.sign(newk2[d][1] + 0.5)
                   v1 = np.dot(trans, get_vshape(fes, k1, newk1[k][0])) 
                   v2 = np.dot(trans, get_vshape(fes, k2, newk2[d][0]))
                   fac = np.sum(v1*v2)/np.sum(v1*v1)*s 
                   #print fac, v1, v2, newk1[k][1], newk2[d][1]
                   #print [newk2[d][2], newk1[k][2]] , fac
                   map[newk2[d][2], newk1[k][2]] = fac
               elif len(d) == 2:
                   dd = np.where(np.sum((pt1 - p)**2, 1) == 0.0)[0]
                   v1 = np.dot(trans, get_vshape(fes, k1, newk1[dd[0]][0])) 
                   v2 = np.dot(trans, get_vshape(fes, k1, newk1[dd[1]][0]))
                   v3 = np.dot(trans, get_vshape(fes, k2, newk2[d[0]][0])) 
                   v4 = np.dot(trans, get_vshape(fes, k2, newk2[d[1]][0]))
                   v1 = v1*np.sign(newk1[dd[0]][1] +0.5)
                   v2 = v2*np.sign(newk1[dd[1]][1] +0.5)
                   v3 = v3*np.sign(newk2[d[0]][1] +0.5)
                   v4 = v4*np.sign(newk2[d[1]][1] +0.5)
                   if (np.abs(np.sum(v1*v3)/np.sum(v1*v1)) == 1
                       and np.abs(np.sum(v2*v4)/np.sum(v2*v2)) == 1):
                       fac1 = np.sum(v1*v3)/np.sum(v1*v1)
                       fac2 = np.sum(v2*v4)/np.sum(v2*v2)

                       #print [newk2[d[0]][2], newk1[dd[0]][2]]
                       #print [newk2[d[1]][2], newk1[dd[1]][2]]
                       map[newk2[d[0]][2], newk1[dd[0]][2]] = fac1
                       map[newk2[d[1]][2], newk1[dd[1]][2]] = fac2
                   elif (np.abs(np.sum(v1*v4)/np.sum(v1*v1)) == 1 and
                         np.abs(np.sum(v2*v3)/np.sum(v2*v2)) == 1):
                       fac1 = np.sum(v1*v4)/np.sum(v1*v1)
                       fac2 = np.sum(v2*v3)/np.sum(v2*v2)
                       #print [newk2[d[1]][2], newk1[dd[0]][2]], fac1
                       #print [newk2[d[0]][2], newk1[dd[1]][2]], fac2
                       map[newk2[d[1]][2], newk1[dd[0]][2]] = fac1
                       map[newk2[d[0]][2], newk1[dd[1]][2]] = fac2
                   else:
                       #print 'two shape vector couples!'
                       m2 = np.transpose(np.vstack((v1, v2)))
                       m1 = np.transpose(np.vstack((v3, v4)))
                       m = np.dot(np.linalg.inv(m1), m2)
                       #print v1, v2, v3, v4, newk1[dd[0]][2], newk1[dd[1]][2],newk2[d[0]][2], newk2[d[1]][2]

                       map[newk2[d[0]][2], newk1[dd[0]][2]] = m[0,0]
                       map[newk2[d[1]][2], newk1[dd[0]][2]] = m[1,0]
                       map[newk2[d[0]][2], newk1[dd[1]][2]] = m[0,1]
                       map[newk2[d[1]][2], newk1[dd[1]][2]] = m[1,1]
               else:
                   pass
           subvdofs1.extend([s for k, v, s in newk1])
           subvdofs2.extend([s for k, v, s in newk2])
       return map

   def find_el_center(fes, ibdr1, trans):
       mesh = fes.GetMesh()
       pts = np.vstack([np.mean([np.dot(trans, mesh.GetVertexArray(kk))
                                 for kk in mesh.GetBdrElementVertices(k)],0)
                        for k in ibdr1])
       return pts
   nbe = fes.GetNBE()
   ibdr1 = [i for i in range(nbe) if fes.GetBdrAttribute(i) == idx1]
   ibdr2 = [i for i in range(nbe) if fes.GetBdrAttribute(i) == idx2]

   ct1 = find_el_center(fes, ibdr1, trans)
   ct2 = find_el_center(fes, ibdr2, trans)

   map_1_2= [np.argmin(np.sum((ct2-c)**2, 1)) for c in ct1]
   ibdr2 = [ibdr2[x] for x in map_1_2]
   ct2 = ct2[map_1_2]

   map = get_all_intpoint(fes, ibdr1, ibdr2, trans)

   return map



'''

MPI utillities

'''
import numpy as np
from mpi4py import MPI
from warnings import warn
from mfem.common.mpi_debug import nicePrint, niceCall

def mpi_type_flag(d):
    if d == MPI.LONG:
        return 1
    elif d == MPI.INT:
        return 2
    elif d == MPI.SHORT:
        return 3
    elif d == MPI.LONG_DOUBLE:
        return 4
    elif d == MPI.DOUBLE:
        return 5
    elif d == MPI.COMPLEX:
        return 6
    elif d == MPI.DOUBLE_COMPLEX:
        return 7
    elif d == MPI.BOOL:
        return 8
    elif d == MPI.FLOAT:
        return 9
    else:
        assert False, "unsupporte data type"


def mpi_type_from_flag(d):
    if d == 1:
        return MPI.LONG
    elif d == 2:
        return MPI.INT
    elif d == 3:
        return MPI.SHORT
    elif d == 4:
        return MPI.LONG_DOUBLE
    elif d == 5:
        return MPI.DOUBLE
    elif d == 6:
        return MPI.COMPLEX
    elif d == 7:
        return MPI.DOUBLE_COMPLEX
    elif d == 8:
        return MPI.BOOL
    elif d == 9:
        return MPI.FLOAT
    assert False, "unsupporte data type"


def mpi_type_to_dtype(mpi_dtype):
    if mpi_dtype == MPI.LONG:
        return np.int64
    if mpi_dtype == MPI.INT:
        return np.int32
    if mpi_dtype == MPI.SHORT:
        return np.int16
    if mpi_dtype == MPI.LONG_DOUBLE:
        return np.float128
    if mpi_dtype == MPI.DOUBLE:
        return np.float64
    if mpi_dtype == MPI.FLOAT:
        return np.float32
    if mpi_dtype == MPI.DOUBLE_COMPLEX:
        return np.complex128
    if mpi_dtype == MPI.COMPLEX:
        return np.complex64
    if mpi_dtype == MPI.BOOL:
        return np.bool
    assert False, "MPI data type is unknown for " + mpi_dtype


def allgather(data):
    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    data = comm.allgather(data)
    MPI.COMM_WORLD.Barrier()
    return data


def allgather_vector(data, mpi_data_type=None):
    from mfem.common.mpi_dtype import get_mpi_datatype
    if mpi_data_type is None:
        mpi_data_type = get_mpi_datatype(data)

    myid = MPI.COMM_WORLD.rank
    rcounts = data.shape[0]
    rcounts = np.array(MPI.COMM_WORLD.allgather(rcounts))

    for x in data.shape[1:]:
        rcounts = rcounts * x
    cm = np.hstack((0, np.cumsum(rcounts)))
    disps = list(cm[:-1])
    length = cm[-1]
    recvbuf = np.empty([length], dtype=data.dtype)
    recvdata = [recvbuf, rcounts, disps, mpi_data_type]
    senddata = [data.flatten(), data.flatten().shape[0]]
    MPI.COMM_WORLD.Allgatherv(senddata, recvdata)
    return recvbuf.reshape(-1, *data.shape[1:])


def gather_vector(data, mpi_data_type=None, root=0):
    '''
    gather vector to root node.
    B: Vector to be collected
    '''
    from mfem.common.mpi_dtype import get_mpi_datatype
    if mpi_data_type is None:
        mpi_data_type = get_mpi_datatype(data)

    myid = MPI.COMM_WORLD.rank
    rcounts = data.shape[0]
    rcounts = MPI.COMM_WORLD.gather(rcounts, root=root)
    cm = np.hstack((0, np.cumsum(rcounts)))
    disps = list(cm[:-1])
    recvdata = None
    senddata = [data, data.shape[0]]

    if myid == root:
        length = cm[-1]
        recvbuf = np.empty([length], dtype=data.dtype)
        recvdata = [recvbuf, rcounts, disps, mpi_data_type]
    else:
        recvdata = [None, rcounts, disps, mpi_data_type]
        recvbuf = None
    MPI.COMM_WORLD.Barrier()
    MPI.COMM_WORLD.Gatherv(senddata, recvdata, root=root)
    if myid == root:
        # print 'collected'
        MPI.COMM_WORLD.Barrier()
        return np.array(recvbuf)
    MPI.COMM_WORLD.Barrier()
    return None


def gather_vector(data, mpi_data_type=None, parent=False,
                  world=MPI.COMM_WORLD, root=0):
    '''
    gather vector to root
    B: Vector to be collected

    for intra-communication, leave parent False. data is gatherd
    to root (myid = 0)

    for inter-communication:
       root group should call with parent = True
       root group should call with data to tell the data type, like np.array(2)
       world should be specified

    '''
    from mfem.common.mpi_dtype import get_mpi_datatype
    myid = world.rank

    if mpi_data_type is None:
        mpi_data_type = get_mpi_datatype(data)

    if world.Is_intra():
        if myid == root:
            parent = True
        rcounts = data.shape[0]
        senddata = [data, data.shape[0]]
    elif parent:
        root = MPI.ROOT if myid == root else MPI.PROC_NULL
        rcounts = 0
        senddata = [np.array(()), 0]
    else:
        rcounts = data.shape[0]
        senddata = [data, data.shape[0]]
        if myid == root:
            parent = True

    rcounts = world.allgather(rcounts)
    cm = np.hstack((0, np.cumsum(rcounts)))
    disps = list(cm[:-1])

    if parent:
        length = cm[-1]
        recvbuf = np.empty([length], dtype=data.dtype)
        recvdata = [recvbuf, rcounts, disps, mpi_data_type]
    else:
        recvdata = [None, rcounts, disps, mpi_data_type]
        recvbuf = None
    world.Barrier()
    world.Gatherv(senddata, recvdata, root=root)
    if parent:
        # print 'collected'
        world.Barrier()
        return np.array(recvbuf)
    else:
        world.Barrier()
        return None


def scatter_vector(vector, mpi_data_type=None, rcounts=None, root=0):
    # scatter data
    #
    # rcounts indicats the amount of data which each process
    # receives
    #
    # for example:     rcounts = fespace.GetTrueVSize()
    if mpi_data_type is None:
        from mfem.common.mpi_dtype import get_mpi_datatype
        if MPI.COMM_WORLD.rank == root:
            mpi_data_type = get_mpi_datatype(vector)
            flag = mpi_type_flag(mpi_data_type)
        else:
            flag = None
        flag = MPI.COMM_WORLD.bcast(flag)
        mpi_data_type = mpi_type_from_flag(flag)

    rcountss = MPI.COMM_WORLD.gather(rcounts, root=root)
    # dprint1(rcountss)
    disps = list(np.hstack((0, np.cumsum(rcountss)))[:-1])
    recvdata = np.empty([rcounts], dtype=mpi_type_to_dtype(mpi_data_type))

    if vector is not None:
        senddata = [vector, rcountss, disps, mpi_data_type]
    else:
        senddata = None

    MPI.COMM_WORLD.Scatterv(senddata, recvdata, root=root)
    # MPI.COMM_WORLD.Barrier()
    return recvdata


'''
def scatter_vector2(vector, mpi_data_type=None, rcounts=None):
    from mfem.common.mpi_dtype import  get_mpi_datatype

    myid     = MPI.COMM_WORLD.rank
    isComplex = check_complex(vector)
    if mpi_data_type is None:
        if myid == 0:
            mpi_data_type = get_mpi_datatype(vector)
            flag = mpi_type_flag(mpi_data_type)
        else:
            flag = None
        flag = MPI.COMM_WORLD.bcast(flag)
        mpi_data_type = mpi_type_from_flag(flag)

    if isComplex:
        if myid == 0:
           r = vector.real +
           i = vector.imag
        else:
           r = None
           i = None
        return   (scatter_vector(r, mpi_data_type, rcounts=rcounts) +
               1j*scatter_vector(i, mpi_data_type, rcounts=rcounts))
    else:
        if myid == 0:
           r = vector
        else:
           r = None
        return scatter_vector(r, mpi_data_type, rcounts = rcounts)
'''


def alltoall_vector(data, datatype):
    '''
    data = list of np.array.
        each array goes to different MPI proc.
        length of data must be equal to MPI size
    '''
    from mfem.common.mpi_dtype import get_mpi_datatype

    num_proc = MPI.COMM_WORLD.size

    senddata = np.hstack(data).astype(datatype, copy=False)
    senddtype = get_mpi_datatype(senddata)
    senddtypes = MPI.COMM_WORLD.allgather(mpi_type_flag(senddtype))
    if len(np.unique(senddtypes)) != 1:
        assert False, "data type is not unique"

    sendsize = np.array([len(x) for x in data], dtype=int)
    senddisp = list(np.hstack((0, np.cumsum(sendsize)))[:-1])

    # (step 1) communicate the size of data
    if len(sendsize) != num_proc:
        assert False, "senddata size does not match with mpi size"
    recvsize = np.empty(num_proc, dtype=int)
    disp = list(range(num_proc))
    counts = [1] * num_proc
    dtype = get_mpi_datatype(recvsize)
    s1 = [sendsize, counts, disp, dtype]
    r1 = [recvsize, counts, disp, dtype]
    MPI.COMM_WORLD.Alltoallv(s1, r1)

    # (step 2) communicate the data
    recvsize = list(recvsize)
    recvdisp = list(np.hstack((0, np.cumsum(recvsize))))
    recvdata = np.empty(np.sum(recvsize), dtype=senddata.dtype)

    s1 = [senddata, sendsize, senddisp, senddtype]
    r1 = [recvdata, recvsize, recvdisp[:-1], senddtype]
    MPI.COMM_WORLD.Alltoallv(s1, r1)

    data = [recvdata[recvdisp[i]:recvdisp[i + 1]] for i in range(num_proc)]
    return data


def alltoall_vectorv(data, datatype):
    '''
    mulitidimension, arbitrary size version
    data = list of list of np.array.
        each list element goes to different MPI proc.
        length of data must be equal to MPI size
    '''
    from mfem.common.mpi_dtype import get_mpi_datatype

    num_proc = MPI.COMM_WORLD.size

    datashape = [np.array([y.shape for y in x]).flatten() for x in data]

    datadim = [np.hstack([len(y.shape) for y in x]) if len(x) > 0 else np.array([], dtype=int)
               for x in data]

    senddata = [np.hstack([y.flatten() for y in x]) if len(x) > 0
                else np.array([], dtype=datatype)
                for x in data]
    sendsize = np.array([len(x) for x in senddata], dtype=int)
    senddisp = list(np.hstack((0, np.cumsum(sendsize)))[:-1])
    senddata = np.hstack(senddata).astype(datatype, copy=False)

    senddtype = get_mpi_datatype(senddata)
    senddtypes = MPI.COMM_WORLD.allgather(mpi_type_flag(senddtype))

    if len(np.unique(senddtypes)) != 1:
        assert False, "data type is not unique"

    # (step 1) communicate the size of data

    datashape = alltoall_vector(datashape, int)
    datadim = alltoall_vector(datadim, int)

    # (step 2) communicate the data
    recvshapes = []
    for dims, shapes in zip(datadim, datashape):
        disps = np.hstack([0, np.cumsum(dims)])
        recvshapes.append([shapes[disps[i]:disps[i + 1]]
                           for i in range(len(disps) - 1)])

    recvsize = [np.sum([np.prod(s) for s in ss]).astype(int)
                for ss in recvshapes]

    recvdisp = list(np.hstack((0, np.cumsum(recvsize))))
    recvdata = np.empty(np.sum(recvsize), dtype=senddata.dtype)

    s1 = [senddata, sendsize, senddisp, senddtype]
    r1 = [recvdata, recvsize, recvdisp[:-1], senddtype]
    MPI.COMM_WORLD.Alltoallv(s1, r1)

    data0 = [recvdata[recvdisp[i]:recvdisp[i + 1]] for i in range(num_proc)]
    data = []
    for d, shapes in zip(data0, recvshapes):
        disps = np.hstack((0, np.cumsum([np.prod(s) for s in shapes])))
        data.append([d[disps[i]:disps[i + 1]].reshape(shapes[i])
                     for i in range(len(shapes))])

    return data


def check_complex(obj, root=0):
    return MPI.COMM_WORLD.bcast(np.iscomplexobj(obj), root=root)


def safe_flatstack(ll, dtype=int):
    if len(ll) > 0:
        return np.hstack(ll).astype(dtype, copy=False)
    else:
        return np.array([], dtype=dtype)


def get_row_partitioning(r_A):
    warn('get_row_partition is deplicated', DeprecationWarning,
         stacklevel=2)
    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank

    m = r_A.GetNumRows()
    m_array = comm.allgather(m)
    rows = [0] + list(np.cumsum(m_array))
    return rows


def get_col_partitioning(r_A):
    warn('get_col_partition is deplicated', DeprecationWarning,
         stacklevel=2)
    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank

    n = r_A.GetNumCols()
    n_array = comm.allgather(n)
    cols = [0] + list(np.cumsum(n_array))
    return cols


def get_partition(A):
    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    return np.linspace(0, A.shape[0], num_proc + 1, dtype=int)


def distribute_vec_from_head(b):
    from mfem.common.mpi_dtype import get_mpi_datatype

    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank

    if myid == 0:
        partitioning = get_partition(b)
        MPItype = get_mpi_datatype(b)
        dtype = b.dtype
        for i in range(num_proc - 1):
            dest = i + 1
            b0 = b[partitioning[dest]:partitioning[dest + 1]]
            comm.send(b0.dtype, dest=dest, tag=dest)
            comm.send(b0.shape, dest=dest, tag=dest)
            comm.Send([b0, MPItype], dest=dest, tag=dest)
        b0 = b[partitioning[0]:partitioning[1]]
    else:
        dtype = comm.recv(source=0, tag=myid)
        shape = comm.recv(source=0, tag=myid)
        b0 = np.zeros(shape, dtype=dtype)
        MPItype = get_mpi_datatype(b0)
        comm.Recv([b0, MPItype], source=0, tag=myid)
    return b0


def distribute_global_coo(A):
    from mfem.common.mpi_dtype import get_mpi_datatype

    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank

    partitioning = get_partition(A)

    row = A.row
    col = A.col
    data = A.data

    ids = np.arange(num_proc)[::-1]

    dtype = data.dtype
    MPItype = get_mpi_datatype(data)

    row2 = []
    col2 = []
    data2 = []

    for i in range(num_proc):
        ids = np.roll(ids, 1)
        pair = ids[myid]

        idx = np.logical_and(
            row >= partitioning[pair], row < partitioning[pair + 1])
        r0 = row[idx].astype(np.int32)
        c0 = col[idx].astype(np.int32)
        d0 = data[idx]

        if pair < myid:  # send first
            comm.send(r0.shape, dest=pair, tag=i)
            comm.Send([r0, MPI.INT], dest=pair, tag=i)
            comm.Send([c0, MPI.INT], dest=pair, tag=i)
            comm.Send([d0, MPItype], dest=pair, tag=i)

            shape = comm.recv(source=pair, tag=i)
            r = np.zeros(shape, dtype=np.int32)
            c = np.zeros(shape, dtype=np.int32)
            d = np.zeros(shape, dtype=dtype)
            comm.Recv([r, MPI.INT], source=pair, tag=i)
            comm.Recv([c, MPI.INT], source=pair, tag=i)
            comm.Recv([d, MPItype], source=pair, tag=i)
        elif pair > myid:  # recv first
            shape = comm.recv(source=pair, tag=i)
            r = np.zeros(shape, dtype=np.int32)
            c = np.zeros(shape, dtype=np.int32)
            d = np.zeros(shape, dtype=dtype)
            comm.Recv([r, MPI.INT], source=pair, tag=i)
            comm.Recv([c, MPI.INT], source=pair, tag=i)
            comm.Recv([d, MPItype], source=pair, tag=i)

            comm.send(r0.shape, dest=pair, tag=i)
            comm.Send([r0, MPI.INT], dest=pair, tag=i)
            comm.Send([c0, MPI.INT], dest=pair, tag=i)
            comm.Send([d0, MPItype], dest=pair, tag=i)
        else:

            r = r0
            c = c0
            d = d0

        row2.append(r)
        col2.append(c)
        data2.append(d)

    from scipy.sparse import coo_matrix

    r = np.hstack(row2) - partitioning[myid]
    c = np.hstack(col2)
    d = np.hstack(data2)

    rsize = partitioning[myid + 1] - partitioning[myid]

    A = coo_matrix((d, (r, c)), shape=(rsize, A.shape[1]),
                   dtype=d.dtype)
    return A



'''
   FormHolder is a block structure to maintain a biuch of lf, bf, and gf.

   Each block can have a multiple bf/lf. So, each element is a dictionary, 
   whose key is a projection operator. 
   Value of dictionary is a two element list. The first one is lf/bf/gf,
   and the second place is for Vector/Matrix 
'''
from itertools import product
import numpy as np

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('FromHolder')


class FormBlock(object):
    def __init__(self, shape, new=None, mixed_new=None, diag=None):
        '''
        kind : scipy
                  stores scipy sparse or numpy array
               hypre
        '''
        try:
            x = len(shape)
        except:
            shape = [shape]

        if len(shape) == 2:
            r, c = shape
            self.block = [[None]*c for x in range(r)]
            self.ndim = 2
        elif len(shape) == 1:
            r = shape[0]
            c = 1
            self.ndim = 1
            self.block = [[None]*c for x in range(r)]
        self._shape = (r, c)
        self._diag = diag
        self.allocator1 = new
        if mixed_new is None:
            self.allocator2 = new
        else:
            self.allocator2 = mixed_new

        self.no_allocation = False

    def __repr__(self):

        text2 = []
        for x in self.block:
            text = []
            for y in x:
                if y is None:
                    text.append("None")
                else:
                    text.append(str(list(y)))
            text2.append(",".join(text))
        full_repr = "\n".join(text2)

        return "Formblock" + str(self._shape) + ":\n" + full_repr
        # return "Formblock" + str(self._shape) + ":\n" + str(self.block)

    @property
    def shape(self):
        return self._shape

    @property
    def diag(self):
        return self._diag

    def set_allocator(self, alloc):
        self.allocator1 = alloc

    def set_mixed_allocator(self, alloc):
        self.allocator2 = alloc

    def set_no_allocator(self):
        self.no_allocation = True
        #self.allocator1 = None
        #self.allocator2 = None

    def __iter__(self):
        assert self.no_allocation, "FormBlock must be fixed"

        all_forms = []
        for r, c in product(range(self.shape[0]), range(self.shape[1])):
            if self.block[r][c] is None:
                continue
            for key in self.block[r][c].keys():
                all_forms.append((r, c, self.block[r][c][key][0]))

        return all_forms.__iter__()

    def __getitem__(self, idx):
        r, c, projector = self.allocate_block(idx)
        if not projector in self.block[r][c]:
            return None

        return self.block[r][c][projector][0]

    def __setitem__(self, idx, v):
        if self.ndim == 2:
            try:
                r, c, projector = idx
            except:
                r, c = idx
                projector = 1
        else:
            c = 0
            try:
                r, projector = idx
            except:
                r = idx
                projector = 1

        if self.block[r][c] is None:
            self.block[r][c] = {}

        self.block[r][c][projector] = [v, None]

    def allocate_block(self, idx, reset=False):
        if self.ndim == 2:
            try:
                r, c, projector = idx
            except:
                r, c = idx
                projector = 1
        else:
            c = 0
            try:
                r, projector = idx
            except:
                r = idx
                projector = 1
        if self.block[r][c] is None:
            self.block[r][c] = {}

        if reset:
            keys = list(self.block[r][c])
            for p in keys:
                del self.block[r][c][p]

        if len(self.block[r][c]) == 0:
            if self.no_allocation and not reset:
                pass
            else:
                if self._diag is None or self._diag[r] == c:
                    form = self.allocator1(r)
                    self.block[r][c][1] = [form, None]
                    projector = 1
                else:
                    form, projector = self.allocator2(r, c)
                    self.block[r][c][projector] = [form, None]
        elif len(self.block[r][c]) == 1:
            projector = list(self.block[r][c])[0]
        else:
            assert False, "should not come here? having two differnt projection in the same block"
        return r, c, projector

    def renew(self, idx):
        self.allocate_block(idx, reset=True)

    def get_projections(self, r, c):
        if self.block[r][c] is None:
            return []
        return list(self.block[r][c])

    def get_matvec(self, r, c=0, p=1):
        return self.block[r][c][p][1]

    def set_matvec(self, r, *args):
        # set_matvec(self, r, c=0, p=1, v):
        if len(args) < 1:
            assert False, "need  a value to set"
        v = args[-1]
        if len(args) == 2:
            c = args[0]
            p = 1
        if len(args) == 3:
            c = args[0]
            p = args[1]
        self.block[r][c][p][1] = v

    def generateMatVec(self, converter1, converter2=None, verbose=False):
        if converter2 is None:
            converter2 = converter1

        for i, j in product(range(self.shape[0]), range(self.shape[1])):
            projs = self.get_projections(i, j)
            for p in projs:
                form = self.block[i][j][p][0]
                if verbose:
                    dprint1("generateMatVec", i, j, form)

                if form is not None:
                    if self._diag is None or self._diag[i] == j:
                        self.set_matvec(i, j, p, converter1(form))
                    else:
                        self.set_matvec(i, j, p, converter2(form))
                else:
                    self.set_matvec(i, j, p, None)


def convertElement(Mreal, Mimag, i, j, converter, projections=None):
    '''
    Generate PyVec/PyMat format data.
    It takes two FormBlocks. One for real and the other for imag.

    M = sum(Opr*proj or proj*Opr), Opr is made from BilinearForms.

    We use the same data structure for lf/gf, for which projector
    will be always 1.
    '''
    keys = set(Mreal.get_projections(i, j) +
               Mimag.get_projections(i, j))

    term = None
    for k in keys:
        if Mreal.block[i][j] is not None:
            rmatvec = Mreal.block[i][j][k][1] if k in Mreal.block[i][j] else None
        else:
            rmatvec = None
        if Mimag.block[i][j] is not None:
            imatvec = Mimag.block[i][j][k][1] if k in Mimag.block[i][j] else None
        else:
            imatvec = None
        m = converter(rmatvec, imatvec)
        if k != 1:
            pos, projector = k
            projections, projections_hash = projections

            h = projections_hash[projector]
            p = projections[h]
            if pos > 0:
                m = m.dot(p)
            else:
                pp = p.transpose()
                m = pp.dot(m)

        if term is None:
            term = m
        else:
            term = term + m
    return term



from __future__ import print_function
from petram.phys.weakform import get_integrators
#
#  a script to produce icon images
#
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc
from matplotlib.transforms import Bbox
import base64
import os

import wx

rc("text", usetex=True)

b64encode = base64.urlsafe_b64encode


def encode(txt):
    return (b64encode(txt.encode())).decode()


# names, domains, ranges, coeffs, dims, wf_forms, strong_forms))
bilinintegs = get_integrators('BilinearOps', return_all=True)
linintegs = get_integrators('LinearOps', return_all=True)


def correct_latex(txt):
    txt = txt.strip()
    txt = txt.replace('\\grad', '\\nabla')
    txt = txt.replace('\\{', '{')
    txt = txt.replace('\\cross', '\\times')
    txt = txt.replace('\\curl', '\\nabla\\times')
    txt = txt.replace('\\div', '\\nabla\\cdot')
    txt = txt.replace('\\ddx', '\\frac{d}{dx}')
    return txt

##################################################


def generate_pix(data, header):

    save_path = os.path.join(os.path.dirname(
        __file__), '..', '..', 'data', 'icon')

    dpi = 72
    dpi2 = dpi*8

    F = plt.figure(num=None, figsize=(3.5, 0.3), dpi=dpi, facecolor='w')
    x = [0, 1, 1, 0, 0]
    y = [1, 1, 0, 0, 1]

    for name, _domain, _range, _coeff, _dim, wf_form, strong_form in data:
        ax = plt.subplot(111)
        ax.cla()
        ax.tick_params(length=0)
        ax.set_axis_off()

        txt1 = correct_latex(wf_form)
        if len(strong_form.strip()) > 0:
            txt2 = correct_latex(strong_form)
            #txt2 = "$\\left[\\equiv "+ txt2[1:-1] + "\\right]$"
            txt2 = "$[\\approx " + txt2[1:-1] + "]$"
            #txt2 = "$\\approxeq "+ txt2[1:-1] + "$"
        else:
            txt2 = ""
        print("text", txt1, txt2)

        plt.text(0.01, 0.3, txt1)

        if txt2 != "":
            plt.text(0.51, 0.3, txt2)

        filename = os.path.join(save_path, header + name + '.png')
        print('filename', filename)
        ed = ax.transAxes.transform([(0, 0), (1, 1)])
        bbox = Bbox.from_extents(
            ed[0, 0]/dpi, ed[0, 1]/dpi, ed[1, 0]/dpi, ed[1, 1]/dpi)
        plt.savefig(filename, dpi=dpi2, format='png', bbox_inches=bbox)

    ax = plt.subplot(111)
    ax.cla()
    ax.tick_params(length=0)
    ax.set_axis_off()

    filename = os.path.join(save_path, header + 'none.png')
    print('filename', filename)
    ed = ax.transAxes.transform([(0, 0), (1, 1)])
    bbox = Bbox.from_extents(
        ed[0, 0]/dpi, ed[0, 1]/dpi, ed[1, 0]/dpi, ed[1, 1]/dpi)
    plt.savefig(filename, dpi=dpi2, format='png', bbox_inches=bbox)

    ##################################################


if __name__ == '__main__':
    generate_pix(bilinintegs, 'form_')
    generate_pix(linintegs, 'form_')



import shutil
import os
import ifigure.widgets.dialog as dialog

def import_ns(model, path = None):
    from ifigure.mto.py_code import PyData
    from ifigure.mto.py_script import PyScript        

    proj = model.get_root_parent()
    if path is None:
        path = dialog.read(wildcard='piScope subtree(pfs)|*.pfs|Python Script (py)|*.py')
        if path == '': return
    if path.endswith(".pfs"):
        if not proj.has_child(name = 'mfem_tmp'):
            proj.add_folder('mfem_tmp')

        child = proj.mfem_tmp.load_subtree(path)

        nsname = '_'.join(child.get_child(idx=0).name.split('_')[:-1])
        nsname0 = nsname
        k = 0
        while model.namespaces.has_child(nsname + '_ns'):
           k = k + 1
           nsname = nsname0 + str(k)

        ns    = child.get_child(name = nsname0+'_ns')
        data  = child.get_child(name = nsname0 + '_data')

        if ns is None: return
        ns.duplicate(model.namespaces, new_name = nsname+'_ns')
        if data is not None:
            data.duplicate(model.datasets, new_name = nsname+'_data')

        proj.mfem_tmp.destroy()
    elif path.endswith('.py'):

        shutil.copy(os.path.join(path), model.namespaces.owndir())
        name = os.path.basename(path)
        sc = model.namespaces.add_childobject(PyScript, name[:-3])
        sc.load_script(os.path.join(model.namespaces.owndir(), name))
        obj =model.datasets.add_data(name[:-6]+'_data')
    else:
        raise AssertionError("Unsupported file")
        
def export_ns(model, name, path = None):
    proj = model.get_root_parent()

    ns    = model.namespaces.get_child(name = name + '_ns')
    data  = model.datasets.get_child(name = name + '_data')

    if ns is None: return
    if path is None:
        path = dialog.write(defaultfile=name+'_mfem_ns.pfs',
                            wildcard='*.pfs')
        if path == '': return
    if not proj.has_child(name = 'mfem_tmp'):
        proj.add_folder('mfem_tmp')
    tmp = proj.mfem_tmp.add_folder(name)
    ns.duplicate(tmp)
    if data is not None:
        data.duplicate(tmp)

    tmp.save_subtree(path)
    proj.mfem_tmp.destroy()



'''
  nodal_refinement

  generate geometrically refined dataset using nodal values.
  note that the data inside the element is simply interplated
  from nodal values and H1 weight. 

'''
import numpy as np
import weakref
import six
from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mfem.par import GlobGeometryRefiner as GR    
else:
    import mfem.ser as mfem
    from mfem.ser import GlobGeometryRefiner as GR
    
Geom = mfem.Geometry()

weight = {}

def surface_weight(refine, gtype):
    if (refine, gtype) in globals()['weight']:
        return globals()['weight'][(refine, gtype)]
    
    quad_v = [[0, 0], [1, 0], [1, 1], [0, 3]]
    quad_e = [[0, 1, 2, 3]]
    tri_v =   [[0, 0], [1, 0], [0, 1]]
    tri_e = [[0, 1, 2,]]
    seg_v =   [[0,], [1, ],]
    seg_e = [[0, 1,]]

    if gtype == mfem.Geometry.TRIANGLE:
        mesh = mfem.Mesh(2, 3, 1, 0, 2)
        for j in range(3):
            mesh.AddVertex(tri_v[j])
        for j in range(1):
            mesh.AddTri(tri_e[j], 11)
        mesh.FinalizeTriMesh(1,1, True)

    elif gtype == mfem.Geometry.SQUARE:
        mesh = mfem.Mesh(2, 4, 1, 0, 2)
        for j in range(4):
            mesh.AddVertex(quad_v[j])
        for j in range(1):
            mesh.AddQuad(quad_e[j], 11)
        mesh.FinalizeQuadMesh(1,1, True)
    elif gtype == mfem.Geometry.SEGMENT:
        mesh = mfem.Mesh(1, 2, 1, 0, 1)
        for j in range(2):
            mesh.AddVertex(seg_v[j])
        for j in range(1):
            seg = mfem.Segment(seg_e[j], j+1)
            mesh.AddElement(seg)
            seg.thisown = False        
            mesh.FinalizeTopology()         
            mesh.Finalize(False, False)

    fec_type = mfem.H1_FECollection
    fe_coll = fec_type(1, 2)
    fes = mfem.FiniteElementSpace(mesh, fe_coll, 2)
    el = fes.GetFE(0)
    npt = Geom.GetVertices(gtype).GetNPoints()
    RefG = GR.Refine(gtype, refine)

    shape = mfem.Vector(el.GetDof())              
    ir = RefG.RefPts

    shapes =[]
    for i in range(ir.GetNPoints()):
        el.CalcShape(ir.IntPoint(i), shape)
        shapes.append(shape.GetDataArray().copy())
        
    w = np.vstack(shapes)
    globals()['weight'][(refine, gtype)] = w

    return w
    

def refine_surface_data(mesh, ibele, val, idx, refine):

    if mesh.Dimension() == 3:
        getarray = mesh.GetBdrArray
        gettrans = mesh.GetBdrElementTransformation
        getelement = mesh.GetBdrElement
        getbasegeom = mesh.GetBdrElementBaseGeometry
        getvertices = mesh.GetBdrElementVertices
    elif mesh.Dimension() == 2:
        getarray = mesh.GetDomainArray
        gettrans = mesh.GetElementTransformation        
        getelement = mesh.GetElement
        getbasegeom = mesh.GetElementBaseGeometry
        getvertices = mesh.GetElementVertices
    else:
        assert False, "BdrNodal Evaluator is not supported for this dimension"


    ptx = []
    data = []
    ridx = []

    nele = 0
    gtype_st = -1    
    sdim = mesh.SpaceDimension()

    for k, i in enumerate(ibele):
        gtype = getbasegeom(i)
        verts = getvertices(i)

        if gtype != gtype_st:
            RefG = GR.Refine(gtype, 3)
            ir = RefG.RefPts
            npt = ir.GetNPoints()
            ele = np.array(RefG.RefGeoms.ToList()).reshape(-1, len(verts))
            w = surface_weight(refine, gtype)
            gtype_st = gtype

        T = gettrans(i)
        pt = np.vstack([T.Transform(ir.IntPoint(j)) for j in range(npt)])
        ptx.append(pt)
        d = val[idx[k]]
        data.append(np.dot(w, d))
        ridx.append(ele + nele)
        nele = nele + ir.GetNPoints()
        
    ptx = np.vstack(ptx)
    data = np.hstack(data)
    ridx = np.vstack(ridx)
    return ptx, data, ridx

def refine_edge_data(mesh, ibele, val, idx, refine):

    if mesh.Dimension() == 2:
        getarray = mesh.GetBdrArray
        gettrans = mesh.GetBdrElementTransformation
        getelement = mesh.GetBdrElement
        getbasegeom = mesh.GetBdrElementBaseGeometry
        getvertices = mesh.GetBdrElementVertices
    elif mesh.Dimension() == 1:
        getarray = mesh.GetDomainArray
        gettrans = mesh.GetElementTransformation        
        getelement = mesh.GetElement
        getbasegeom = mesh.GetElementBaseGeometry
        getvertices = mesh.GetElementVertices
    else:
        assert False, "BdrEdge Evaluator is not supported for this dimension"


    ptx = []
    data = []
    ridx = []

    nele = 0
    gtype_st = -1    
    sdim = mesh.SpaceDimension()

    for k, i in enumerate(ibele):
        gtype = getbasegeom(i)
        verts = getvertices(i)

        if gtype != gtype_st:
            RefG = GR.Refine(gtype, refine)
            ir = RefG.RefPts
            npt = ir.GetNPoints()
            ele = np.array(RefG.RefGeoms.ToList()).reshape(-1, len(verts))
            w = surface_weight(refine, gtype)
            gtype_st = gtype

        T = gettrans(i)
        pt = np.vstack([T.Transform(ir.IntPoint(j)) for j in range(npt)])
        ptx.append(pt)
        d = val[idx[k]]
        data.append(np.dot(w, d))
        ridx.append(ele + nele)
        nele = nele + ir.GetNPoints()
        
    ptx = np.vstack(ptx)
    data = np.hstack(data)
    ridx = np.vstack(ridx)
    return ptx, data, ridx




import os
import sys
import wx
import traceback
import numpy as np
import weakref

import ifigure.widgets.dialog as dialog
import ifigure.events

from ifigure.widgets.book_viewer import FramePlus
from ifigure.utils.cbook import BuildMenu
from petram.mfem_viewer import MFEM_menus

class SimpleFramePlus(FramePlus):
    def __init__(self, parent, *args, **kwargs):
        self.close_cb = kwargs.pop("close_cb", None)

        if "wxMac" not in wx.PlatformInfo:        
            kwargs['nomenu'] = True
            
        super(SimpleFramePlus, self).__init__(parent, *args, **kwargs)

        self.Bind(wx.EVT_CLOSE, self.onClose)
        wx.GetApp().add_palette(self)
        self._atable = []
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F2, wx.ID_BACKWARD))
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F1, wx.ID_FORWARD))
        atable = wx.AcceleratorTable(self._atable)
        self.SetAcceleratorTable(atable)

        # need to remove self from table
        tw = wx.GetApp().TopWindow
        tw.windowlist.remove_item(self)

        #self.Bind(wx.EVT_MENU, lambda evt: frame.ProcessEvent(evt))
        if "wxMac" in wx.PlatformInfo:
            extra_menu = wx.Menu()
            self.menuBar.Insert(self.menuBar.GetMenuCount(), extra_menu,"MFEM")
            menus = MFEM_menus(parent)
            ret = BuildMenu(extra_menu, menus)

            self.SetMenuBar(self.menuBar)
        else:
#            self.SetMenuBar(self.menuBar)
            self.SetMenuBar(None)

    def onResize(self, evt):
        evt.Skip()

    def onActivate(self, evt):
        pass

    def onUpdateUI(self, evt):
        pass

    def onClose(self, evt):
        wx.GetApp().rm_palette(self)
        if self.close_cb is not None:
            self.close_cb(evt)
        self.Destroy()
        evt.Skip()



'''
   extract_refined_mesh_data3

   extract refined  mesh data from 2D (=ndim)  mesh
'''
import numpy as np
from collections import defaultdict

from petram.mesh.find_edges import find_edges
from petram.mesh.find_vertex import find_vertex
from mfem.ser import GlobGeometryRefiner as GR

def extract_refined_mesh_data1(mesh, refine=None):
    ndim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    ivert0 = [mesh.GetElement(i).GetVerticesArray()
              for i in range(mesh.GetNE())]
    idx2 = np.arange(mesh.GetNE())
    iv2 = [np.arange(mesh.GetNV())]

    table = np.arange(mesh.GetNV())

    battrs = mesh.GetAttributeArray()
    gt = mesh.GetElementTransformation

    cells = {}
    cell_data = {}

    ptx = [mesh.GetVertexArray() for i in range(mesh.GetNV())]

    base = 1

    from petram.mesh.refined_mfem_geom import get_geom

    if len(idx2) > 0:
        attr2, ptx2, ivx2, ivxe2, attrx2 = get_geom(idx2, 2, base, gt, battrs,
                                                    sdim, refine,)
        ptx.append(ptx2)
        cells['line_x'] = ivx2
        cell_data['line_x'] = {}
        cell_data['line_x']['physical'] = attrx2
        cell_data['X_refined_edge']=ptx2

    X = np.vstack(ptx)

    from petram.mesh.mesh_utils import populate_plotdata

    l_s_loop = populate_plotdata(mesh, table, cells, cell_data)

    iedge2bb = None # is it used?

    #print "X", X.shape
    #for k in cells['vertex']:print X[k]
    return X, cells, cell_data, l_s_loop, iedge2bb





def select_sol(solname = ''):
    if not model.solutions.has_child(solname): return

    sol = model.solutions.get_child(name = solname)
    model.param.setvar(sol, '='+sol.get_full_path())

    model.scripts.helpers.read_sol()
    return sol
ans(select_sol(*args, **kwargs))



from __future__ import print_function
from ifigure.widgets.canvas.ifigure_canvas import ifigure_canvas
'''
   Viewer/Editor of MFEM model

   * this modules is designed to avoid importing mfem directly here

'''
import numpy as np
import os
import wx
import traceback
import weakref
from ifigure.interactive import figure
from ifigure.widgets.book_viewer import BookViewer
from ifigure.utils.cbook import BuildMenu
import ifigure.widgets.dialog as dialog
import ifigure.events
from petram.pi.dlg_progressbar import progressbar

try:
    import petram.geom
    hasGeom = True
except ImportError:
    hasGeom = False


def setup_figure(fig):
    fig.nsec(1)
    fig.property(fig.get_axes(0), 'axis', False)
    fig.get_page(0).set_nomargin(True)
    fig.property(fig.get_page(0), 'bgcolor', 'white')


ID_SOL_FOLDER = wx.NewIdRef(count=1)


def MFEM_menus(parent):
    self = parent
    menu1 = [("+Open Model...", None, None),
             ("Binary...", self.onOpenPMFEM, None, None),
             ("Script/Data Files...", self.onOpenModelS, None),
             ("!", None, None),
             ("+Mesh", None, None),
             ("New Mesh File...",  self.onNewMesh, None),

             ("Reload Mesh",  self.onLoadMesh, None),
             ("Mesh visualization +",  self.onMeshVisPlus, None),
             ("Mesh visualization -",  self.onMeshVisMinus, None),
             ("!", None, None),
             ("+Namespace", None, None),
             ("New...", self.onNewNS, None),
             ("Load...", self.onLoadNS, None),
             ("Export...", self.onExportNS, None),
             ("Rebuild", self.onRebuildNS, None),
             ("!", None, None),
             ("Edit Model...", self.onEditModel, None),
             ("+Solve", None, None),
             ("Serial",    self.onSerDriver, None),
             ("Parallel",  self.onParDriver, None),
             ("+Extra", None, None),
             ("+Store solution to", None, None, None, ID_SOL_FOLDER),
             ("!", None, None),
             ("---", None, None),
             ("New sol...",   self.onNewLocalSol, None),
             ("Clear sol", self.onClearSol, None),
             ("Preprocess data",   self.onRunPreprocess, None),
             ("!", None, None),
             ("!", None, None), ]

    if petram.mfem_model.has_cluster_access:
        menu2 = [("+Cluster", None, None),
                 ("Setting...", self.onServerSetting, None),
                 ("Solve...", self.onServerSolve, None),
                 ("Retrieve File", self.onServerRetrieve, None),
                 ("!", None, None), ]
    else:
        menu2 = []

    menu3 = [("+Plot", None, None),
             ("Function...",    self.onPlotExpr, None),
             ("Solution ...",    self.onDlgPlotSol, None),
             ("!", None, None),
             #("+Solution", None, None, None, ID_SOL_FOLDER),
             #("Reload Sol", None, None,),
             #("Clear...",    self.onClearSol, None),
             #("!", None, None),
             ("+Export Model...", self.onSaveModel, None),
             ("Binary...", self.onSaveModel, None),
             ("Script/Data Files...", self.onSaveModelS, None),
             ("!", None, None),
             ("---", None, None),
             ("Reset Model", self.onResetModel, None), ]
    return menu1 + menu2 + menu3


class MFEMViewerCanvas(ifigure_canvas):
    def __init__(self, *args, **kwags):
        self._mfemviewer_spacehit = False
        return ifigure_canvas.__init__(self, *args, **kwags)

    def unselect_all(self):
        ifigure_canvas.unselect_all(self)
        self.TopLevelParent._dom_bdr_sel = ([], [], [], [])

    def onKey(self, evt):
        keycode = evt.guiEvent.GetKeyCode()
        if keycode == wx.WXK_SPACE:
            self._mfemviewer_spacehit = True

        return ifigure_canvas.onKey(self, evt)

    def onKey2(self, evt):
        keycode = evt.guiEvent.GetKeyCode()
        if keycode == wx.WXK_SPACE:
            if self._mfemviewer_spacehit:
                self.TopLevelParent.handle_shifthit()
        self._mfemviewer_spacehit = False
        return ifigure_canvas.onKey2(self, evt)


class MFEMViewer(BookViewer):
    def __init__(self, *args, **kargs):
        kargs['isattachable'] = False
        kargs['isinteractivetarget'] = False
        kargs['ismultipage'] = False
        BookViewer.__init__(self, *args, **kargs)
        extra_menu = wx.Menu()
        self.menuBar.Insert(self.menuBar.GetMenuCount()-1,
                            extra_menu, "PetraM")
        menus = MFEM_menus(self)
        ret = BuildMenu(extra_menu, menus)

        data = ret[ID_SOL_FOLDER]
        if len(data) == 2:
            self._solmenu, item = data
            self._ID_SOL_FOLDER = item.GetId()
        else:
            self._solmenu = data
            self._ID_SOL_FOLDER = ID_SOL_FOLDER

        self._hidemesh = True
        self._sel_mode = ''  # selecting particular geomgetry element
        self._view_mode = ''  # ('geom', 'mesh', 'phys')
        self._s_v_loop = {}
        self._selected_volume = []    # store selected volume
        self._figure_data = {}

        # extra refinement of mesh for visualization
        self._mesh_vis = 0
        # hidden element in MFEM mode
        self._hidden_volume = []

        # hidden element in mesh mode
        # for mesh we need to keep track face and edge too.
        self._mhidden_volume = []
        self._mhidden_face = []
        self._mhidden_edge = []

        self._view_mode_group = ''
        self._is_mfem_geom_fig = False
        self._dom_bdr_sel = ([], [], [], [])
        self._palette_focus = ''
        self.model = self.book.get_parent()
        self.editdlg = None
        self.selection_palette = None
        self.geom_info_palette = None
        self.plotsoldlg = None
        self.plotexprdlg = None
        self.engine = None
        self.dombdr = None

        from petram.pi.sel_buttons import btask, refresh

        self.install_toolbar_palette('petram_phys',
                                     btask,
                                     mode='3D',
                                     refresh=refresh)
        from petram.mesh.mesh_sel_buttons import btask
        self.install_toolbar_palette('petram_mesh',
                                     btask,
                                     mode='3D',
                                     refresh=refresh)
        self.use_toolbar_palette('petram_phys',
                                 mode='3D')
        if hasGeom:
            from petram.geom.geom_sel_buttons import btask
            self.install_toolbar_palette('petram_geom',
                                         btask,
                                         mode='3D',
                                         refresh=refresh)

        od = self.model.param.getvar('mfem_model')
        od.set_root_path(self.model.owndir())

        if od is None:
            self.model.scripts.helpers.reset_model()

        self.cla()
        setup_figure(self)
        self.start_engine()

        for child in od.walk():
            view_mode, name, data = child.load_gui_figure_data(self)
            if data is not None:
                self.set_figure_data(view_mode, name, data)

        self.model.scripts.helpers.rebuild_ns()

        if self.model.variables.getvar('mesh') is None:
            try:
                self.load_mesh()
                self.engine.run_mesh_extension_prep(reset=True)
            except:
                dialog.showtraceback(parent=self,
                                     txt='mesh file load error',
                                     title='Error',
                                     traceback=traceback.format_exc())
        self.plot_mfem_geom()

        self.Bind(ifigure.events.TD_EVT_ARTIST_DRAGSELECTION,
                  self.onTD_DragSelectionInFigure)

        self.canvas._popup_style = 1  # popup_skip_2d
        self.canvas.__class__ = MFEMViewerCanvas

        # make an empty sol folder if it is not defined.
        if self.model.param.eval('sol') is None:
            self.model.scripts.helpers.make_new_sol()

        import logging
        numba_logger = logging.getLogger('numba')
        numba_logger.setLevel(logging.WARNING)
        print("numba debug logging is suppressed")

    @property
    def view_mode_group(self):
        return self._view_mode_group

    @property
    def dom_bdr_sel(self):
        return self._dom_bdr_sel

    def set_view_mode(self, mode, mm=None):
        p = mm
        while p is not None:
            if p.is_viewmode_grouphead():
                break
            p = p.parent
        group = p.name() if p is not None else 'root'

        do_palette = (self._view_mode != mode)
        do_plot = (self._view_mode != mode) or (self._view_mode_group != group)

        self._view_mode = mode
        self._view_mode_group = p.name() if p is not None else 'root'

        if do_palette:
            if mode == '':
                self.use_toolbar_std_palette()
            else:
                self.use_toolbar_palette('petram_'+mode, mode='3D')
        if do_plot:
            if p is not None:
                # update figure data
                p.update_figure_data(self)
                #print("calling do_plot", self._view_mode, p.figure_data_name())
                self.update_figure(self._view_mode, p.figure_data_name(),
                                   updateall=True)

    def set_figure_data(self, view_mode, name, data):
        if not view_mode in self._figure_data:
            self._figure_data[view_mode] = {}
        self._figure_data[view_mode][name] = data

    def del_figure_data(self, view_mode, name):
        if not view_mode in self._figure_data:
            self._figure_data[view_mode] = {}
        if name in self._figure_data[view_mode]:
            del self._figure_data[view_mode][name]

    def update_figure(self, view_mode, name, updateall=False):
        from petram.mesh.geo_plot import plot_geometry, oplot_meshed
        from petram.mesh.geo_plot import hide_face_meshmode, hide_edge_meshmode

        if self._is_mfem_geom_fig and view_mode == 'phys':
            self.onHideMesh()
        elif self._is_mfem_geom_fig and view_mode == 'mesh' and name == 'mfem':
            self.onShowMesh()
        else:
            self._is_mfem_geom_fig = False
            if not view_mode in self._figure_data:
                self.cls()
                return
            if view_mode == 'geom':
                d = self._figure_data['geom']
                self._hidden_volume = []
                if name in d:
                    ret = d[name]
                    plot_geometry(self,  ret)
                else:
                    #print('Geometry figure data not found :' + name)
                    self.cls()
                    return

            elif view_mode == 'mesh':
                if name == 'mfem':
                    if not 'mfem' in self._figure_data['mesh']:
                        self.cls()
                    else:
                        d = self._figure_data['mesh']['mfem']
                        plot_geometry(self, d, geo_phys='physical', lw=1.0)
                        self._is_mfem_geom_fig = True
                else:
                    d = self._figure_data['mesh']
                    if updateall or not name[0] in d:
                        if not 'geom' in self._figure_data:
                            # geom is not yet run
                            self.cls()
                            return
                        d = self._figure_data['geom']
                        plot_geometry(self,  d[name[1]])
                    if name[0] in d:
                        #print("calling oplot")
                        oplot_meshed(self,  d[name[0]])
                        self._hidemesh = False
                    else:
                        hide_face_meshmode(self, [])
                        hide_edge_meshmode(self, [])

            elif view_mode == 'phys':
                ret = self._figure_data['phys']
                plot_geometry(self,  ret, geo_phys='physical')
                self._is_mfem_geom_fig = True
                self._hidemesh = True

    def onUpdateUI(self, evt):
        if evt.GetId() == self._ID_SOL_FOLDER:
            m = self._solmenu
            for item in m.GetMenuItems():
                m.DestroyItem(item)
            try:
                if not self.model.solutions.has_owndir():
                    return
                dir = self.model.solutions.owndir()
                sol_names = [x for x in os.listdir(
                    dir) if os.path.isdir(os.path.join(dir, x))]
                sol_names = sorted(sol_names)
                menus = []
                for n in sol_names:
                    menus.append((n, n))
            except:
                import traceback
                traceback.print_exc()
                evt.Enable(False)
                return
            mm = []
            from petram.sol.solsets import read_solsets, find_solfiles
            for m0, m2 in menus:
                def handler(evt, dir0=m0):
                    # self.model.scripts.helpers.rebuild_ns()
                    # self.engine.assign_sel_index()
                    #path = os.path.join(dir, dir0)
                    #print('loading sol from ' + path)
                    model = self.model
                    folder = model.solutions.get_child(name=str(dir0))
                    param = model.param
                    param.setvar('sol', '='+folder.get_full_path())
                    m = self.model.param.getvar('mfem_model')
                    m.set_root_path(self.model.owndir())
                    evt.Skip()

                mm.append((m2, 'Store solution in ' + m0, handler))

            # mm.append(('Other...', 'Load from ohter place (FileDialog will open)',
            #           handler2))
            param = self.model.param
            sol = param.eval('sol')
            choice = sol.name if sol is not None else None

            if len(mm) > 0:
                for a, b, c in mm:
                    mmi = self.add_menu(m, wx.ID_ANY, a, b,
                                        c, kind=wx.ITEM_CHECK)
                    if a == choice:
                        mmi.Check(True)
            evt.Enable(True)
        else:
            super(MFEMViewer, self).onUpdateUI(evt)

    def start_engine(self):
        self.engine = self.model.scripts.helpers.start_engine()
        # if self.model.variables.hasvar('engine')
        #from engine import SerialEngine
        #self.engine = SerialEngine()
        #self.model.variables.setvar('engine', self.engine)

    def onOpenPMFEM(self, evt):
        import petram.helper.pickle_wrapper as pickle
        from ifigure.mto.py_code import PyData
        from ifigure.mto.py_script import PyScript

        from petram.mesh.mesh_model import MeshFile
        path = dialog.read(
            message='Select model file to read', wildcard='*.pmfm')
        if path == '':
            return

        od = pickle.load(open(path, 'rb'))
        self.model.param.setvar('mfem_model', od)

        # clean up namespaces/datasets
        for name, child in self.model.namespaces.get_children():
            child.destroy()
        for name, child in self.model.datasets.get_children():
            child.destroy()

        # first expand all namelist under namespaces
        dir = self.model.namespaces.owndir()

        ns_names = []
        for node in od.walk():
            if node.has_ns():
                if node.ns_name is None:
                    continue
                if not node.ns_name in ns_names:
                    ns_names.append(node.ns_name)
                    node.write_ns_script_data(dir=dir)

        for file in os.listdir(dir):
            print("processing file", file)
            if file == os.path.basename(path):
                continue
            if file.endswith('.py'):
                #shutil.copy(os.path.join(dir, file), self.model.namespaces.owndir())
                sc = self.model.namespaces.add_childobject(PyScript, file[:-3])
                sc.load_script(os.path.join(
                    self.model.namespaces.owndir(), file))
            if file.endswith('.dat'):
                fid = open(os.path.join(dir, file), 'rb')
                data = pickle.load(fid)
                fid.close()
                obj = self.model.datasets.add_childobject(
                    PyData, file[:-6]+'data')
                obj.setvar(data)

        self.cla()
        self.load_mesh()

        if self.editdlg is not None:
            od = self.model.param.getvar('mfem_model')
            self.editdlg.set_model(od)
        self.model.variables.setvar('modelfile_path', path)
        evt.Skip()

    def onOpenModelS(self, evt):
        import imp
        import shutil
        import ifigure.utils.pickle_wrapper as pickle
        from ifigure.mto.py_code import PyData
        from ifigure.mto.py_script import PyScript
        path = dialog.read(
            message='Select model file to read', wildcard='*.py')
        try:
            print("loading", path)
            m = imp.load_source('petram.user_model', path)
            model = m.make_model()
        except:
            print(path)
            dialog.showtraceback(parent=self,
                                 txt='Model file load error',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return
        dir = os.path.dirname(path)
        c = [child for name, child in self.model.datasets.get_children()]
        for child in c:
            child.destroy()
        c = [child for name, child in self.model.namespaces.get_children()]
        for child in c:
            child.destroy()

        for file in os.listdir(dir):
            if file == os.path.basename(path):
                continue
            if file.endswith('.py'):
                shutil.copy(os.path.join(dir, file),
                            self.model.namespaces.owndir())
                sc = self.model.namespaces.add_childobject(PyScript, file[:-3])
                sc.load_script(os.path.join(
                    self.model.namespaces.owndir(), file))
            if file.endswith('.dat'):
                fid = open(os.path.join(dir, file), 'rb')
                data = pickle.load(fid)
                fid.close()
                obj = self.model.datasets.add_childobject(
                    PyData, file[:-6]+'data')
                obj.setvar(data)

        self.model.param.setvar('mfem_model', model)
        self.cla()
        self.load_mesh()

        if self.editdlg is not None:
            od = self.model.param.getvar('mfem_model')
            self.editdlg.set_model(od)

        evt.Skip()

    def _getSelectedIndexVolume(self, already_selected_surface):
        objs = [o().figobj for o in self.canvas.selection
                if o() is not None and o().figobj.name.startswith('face')]

        unselected = []
        selected = []
        for o in objs:
            idx = set(o.getSelectedIndex())
            unselected.extend(list(already_selected_surface.difference(idx)))
            selected.extend(list(idx.difference(already_selected_surface)))

        unselected = list(set(unselected))
        selected = list(set(selected))

        return selected, unselected, objs

    def _getSelectedIndex(self, mode='face'):
        objs = [o().figobj for o in self.canvas.selection
                if o() is not None]

        sel = []
        oo = []
        for obj in objs:
            if obj.name.startswith(mode):
                sel.extend(obj.getSelectedIndex())
                oo.append(obj)
        return sel, oo

    def _getFigObjs(self, mode='face'):
        objs = []
        ax = self.get_axes()
        for name, obj in ax.get_children():
            if name.startswith(mode):
                objs.append(obj)
        return objs

    def onTD_DragSelectionInFigure(self, evt):
        # print("onTD_DragSelectionInFigure")
        self.onTD_SelectionInFigure(evt)

    def onTD_SelectionInFigure(self, evt=None):
        if len(self.canvas.selection) == 0:
            self._dom_bdr_sel = ([], [], [], [])
        #    return

        status_txt = ''
        if not self._view_mode in self._s_v_loop:
            self.set_status_text('', timeout=600000)
            evt.selections = self.canvas.selection
            self.property_editor.onTD_Selection(evt)
            return

        #print("canvas sel",  self.canvas.selection)
        _s_v_loop = self._s_v_loop[self._view_mode]
        sf, sv, se, sp = [], [], [], []
        if self._sel_mode == 'volume':
            if _s_v_loop[1] is None:
                return

            sl = _s_v_loop[1]
            already_selected_surface = set(
                sum([sl[k] for k in self._dom_bdr_sel[0]], []))

            already_selected = self._dom_bdr_sel[0]
            already_selected = [k for k in already_selected if k in sl]

            selected, unselected, objs = self._getSelectedIndexVolume(
                already_selected_surface)

            selected_volume = already_selected[:]

            if len(unselected) > 0:
                for i in unselected:
                    for k in sl.keys():
                        if i in sl[k] and k in selected_volume:
                            selected_volume.remove(k)
            else:
                for i in selected:
                    for k in sl.keys():
                        if i in sl[k]:
                            selected_volume.append(k)

            selected_volume = list(set(selected_volume))

            hidden = self._mhidden_volume if self._view_mode == 'mesh' else self._hidden_volume
            hidden = list(set(hidden))

            for x in hidden:
                if x in selected_volume:
                    selected_volume.remove(x)

            status_txt = 'Volume :' + \
                ','.join([str(x) for x in selected_volume])

            surf_idx = []
            for kk in selected_volume:
                surf_idx.extend(sl[kk])
            surf_idx = list(set(surf_idx))

            objs = self._getFigObjs(mode='face')

            for o in objs:
                o.setSelectedIndex(surf_idx)
                if len(surf_idx) > 0:
                    if len(self.canvas.selection) > 0 and len(o._artists) > 0:
                        self.canvas.add_selection(o._artists[0])
                else:
                    self.canvas.unselect_all()
            sv = selected_volume

        elif self._sel_mode == 'face':
            idx, objs = self._getSelectedIndex(mode='face')
            status_txt = 'Face: ' + ','.join([str(x) for x in idx])
            v = _s_v_loop[1]
            connected_vol = []
            if v is not None:  # in case surface loop is defined (3D)

                for i in idx:
                    for k in v.keys():
                        if i in v[k]:
                            connected_vol.append(k)
                connected_vol = list(set(connected_vol))
                status_txt += ('(Volume: ' +
                               ','.join([str(x) for x in connected_vol]) + ')')
            sv = connected_vol
            sf = idx

        elif self._sel_mode == 'edge':
            idx, objs = self._getSelectedIndex(mode='edge')
            s = _s_v_loop[0]
            connected_surf = []
            if s is not None:  # in case line loop is defined...(2D/3D)
                for i in idx:
                    for k in s.keys():
                        if i in s[k]:
                            connected_surf.append(k)
                connected_suf = list(set(connected_surf))
                sf = connected_suf
                status_txt = ('Edge: ' + ','.join([str(x) for x in idx]) + '(Face: ' +
                              ','.join([str(x) for x in connected_suf]) + ')')
            else:
                status_txt = 'Edge: ' + ','.join([str(x) for x in idx])

            se = idx

        elif self._sel_mode == 'point':
            if (self.get_axes().has_child('point') and
                    not self.get_axes().point.isempty()):
                point = self.get_axes().point
                idx = point.getSelectedIndex()
                aidx = point.getvar('array_idx')

                status_txt = 'Vertex: ' + ','.join([str(x) for x in idx])
                if len(idx) == 1:
                    ii = np.where(aidx == idx[0])[0][0]
                    t = (" (" + str(point.getvar('x')[ii]) + ", " +
                         str(point.getvar('y')[ii]) + ", " +
                         str(point.getvar('z')[ii]) + ")")
                    status_txt = status_txt + t
                elif len(idx) == 2:
                    x, y, z = point.getvar('x', 'y', 'z')
                    ii1 = np.where(aidx == idx[0])[0][0]
                    ii2 = np.where(aidx == idx[1])[0][0]
                    dx = x[ii1] - x[ii2]
                    dy = y[ii1] - y[ii2]
                    dz = z[ii1] - z[ii2]
                    dd = np.sqrt(dx**2+dy**2+dz**2)
                    t = (" (delta = " + str(dx) + ", " + str(dy) + ", "+str(dz) +
                         ", dist. = " + str(dd))
                    status_txt = status_txt + t
                else:
                    pass
            else:
                idx = []
            sp = idx
        else:
            pass

        self.set_status_text(status_txt, timeout=600000)

        self._dom_bdr_sel = (sv, sf, se, sp)
        evt.selections = self.canvas.selection
        self.property_editor.onTD_Selection(evt)

    def onNewMesh(self, evt):
        from ifigure.widgets.dialog import read
        from petram.mesh.mesh_model import MeshFile, MFEMMesh
        path = read(message='Select mesh file to read',
                    wildcard='MFEM|*.mesh|Gmsh|*.msh')
        if path == '':
            return
        od = self.model.param.getvar('mfem_model')

        mg = MFEMMesh()
        data = MeshFile(path=path)

        nameg = od['Mesh'].add_itemobj('MFEMMesh', mg)
        name = od['Mesh'][nameg].add_itemobj('MeshFile', data)
        for key in od['Mesh']:
            if key != nameg:
                od['Mesh'][key].enabled = False
        self.load_mesh()

    def onLoadMesh(self, evt):
        win = evt.GetEventObject()
        if hasattr(win, 'GetTopLevelParent'):
            top_win = win.GetTopLevelParent()
        else:
            top_win = self
        dlg = dialog.progressbar(top_win, 'Loading mesh...',
                                 'In progress', 5)
        dlg.Show()
        wx.GetApp().Yield()

        self.load_mesh()
        self._hidemesh = True

        dlg.Destroy()

    def load_mesh(self):
        if self.engine is None:
            self.start_engine()
        od = self.model.param.getvar('mfem_model')

        import wx
        projfile = wx.GetApp().TopWindow.proj.getvar('filename')
        if projfile is not None:
            od.model_path = os.path.dirname(projfile)
        self.engine.set_model(od)
        try:
            cdir = os.getcwd()
        except:
            from os.path import expanduser
            cdir = expanduser("~")

        try:
            err = -1
            os.chdir(self.model.owndir())
            try:
                self.engine.run_mesh(skip_refine=True)
            except:
                exception = traceback.format_exc()
                assert False, "error in run_mesh"
            mesh = self.engine.get_mesh()
            self.model.variables.setvar('mesh', mesh)
            os.chdir(cdir)

        except:
            os.chdir(cdir)
            dialog.showtraceback(parent=self,
                                 txt='Mesh load error',
                                 title='Error (Load Mesh)',
                                 traceback=exception)
        try:
            err, exception = self.engine.run_config()
            if err != 0:
                assert False, "error in run_config"
        except:
            dialog.showtraceback(parent=self,
                                 txt='Error during run_config after loading mesh',
                                 title='Error (run_config)',
                                 traceback=exception)

        if err != -1:
            self.plot_mfem_geom()
            self.use_toolbar_palette('petram_phys', mode='3D')

    def onMeshVisPlus(self, evt):
        self._mesh_vis = self._mesh_vis + 1
        self.onLoadMesh(evt)

    def onMeshVisMinus(self, evt):
        self._mesh_vis = self._mesh_vis - 1
        self.onLoadMesh(evt)

    def plot_mfem_geom(self):
        from petram.mesh.geo_plot import plot_geometry
        from petram.mesh.read_mfemmesh import extract_mesh_data

        mesh = self.model.variables.getvar('mesh')
        if mesh is not None:
            from petram.mesh.refined_mfem_geom import default_refine as refine

            if mesh.GetNodalFESpace() is not None:
                refine = mesh.GetNodalFESpace().GetOrder(0)
            else:
                refine = 1
            refine = refine + self._mesh_vis
            X, cells, cell_data, sl, iedge2bb = extract_mesh_data(mesh, refine)
            self._s_v_loop['phys'] = sl
            self._s_v_loop['mesh'] = sl
            ret = (X, cells, None, cell_data, None)
            self._ret_ret = ret, iedge2bb
            plot_geometry(self, ret, geo_phys='physical')
            self._figure_data['phys'] = ret
            if not 'mesh' in self._figure_data:
                self._figure_data['mesh'] = {}
            self._figure_data['mesh']['mfem'] = ret
            self._view_mode = 'phys'
            self._is_mfem_geom_fig = True
        else:
            self.cls()

    def highlight_element(self, sel):
        if not self._view_mode in self._s_v_loop:
            return None

        _s_v_loop = self._s_v_loop[self._view_mode]
        ax = self.get_axes()

        if len(sel['volume']) != 0:
            if _s_v_loop[1] is None:
                return None
            sl = _s_v_loop[1]
            vfaces = []
            for i in sel['volume']:
                if i in sl:
                    vfaces.extend(sl[i])
                else:
                    print('Volume: ' + str(i) + " not found")
            vfaces = list(set(vfaces))
        else:
            vfaces = []

        obj = None
        robj = []
        for name, obj in ax.get_children():
            if name.startswith('point'):
                obj.setSelectedIndex(sel['point'])
                if len(sel['point']) != 0:
                    robj.append(obj)

            if name.startswith('edge'):
                obj.setSelectedIndex(sel['edge'])
                if len(sel['edge']) != 0:
                    robj.append(obj)

            if name.startswith('face'):
                obj.setSelectedIndex(sel['face'])
                if len(sel['face']) != 0:
                    robj.append(obj)
                if len(vfaces) > 0:
                    obj.setSelectedIndex(vfaces)
                    robj.append(obj)

        return robj

    def highlight_domain(self, i):
        '''
        i is 1-based index
        '''
        if not self._view_mode in self._s_v_loop:
            return

        _s_v_loop = self._s_v_loop[self._view_mode]
        ax = self.get_axes()
        try:
            x = len(i)
        except:
            i = list(i)

        hidden = (self._mhidden_volume if self._view_mode == 'mesh'
                  else self._hidden_volume)
        hidden = list(set(hidden))

        i = [x for x in i if not x in hidden]

        self.canvas.unselect_all()
        for name, obj in ax.get_children():
            if not name.startswith('face'):
                continue
            if _s_v_loop[1] is None:
                return
            if len(i) > 0:
                sl = _s_v_loop[1]
                faces = []
                for key in i:
                    if key in sl:
                        faces.extend(sl[key])
                    else:
                        print('Volume: ' + str(key) + " not found")
                obj.setSelectedIndex(faces)
                if len(obj._artists) > 0:
                    self.canvas.add_selection(obj._artists[0])
            else:
                obj.setSelectedIndex([])
        wx.CallAfter(self.canvas.refresh_hl)

    def highlight_face(self, i):
        '''
        i is 1-based index
        '''
        try:
            x = len(i)
        except:
            i = list(i)
        ax = self.get_axes()

        self.canvas.unselect_all()

        for name, obj in ax.get_children():
            if not name.startswith('face'):
                continue
            if len(i) > 0:
                obj.setSelectedIndex(i)
                #print("add_selection", obj, obj._artists[0])
                if len(obj._artists) > 0:
                    self.canvas.add_selection(obj._artists[0])
            else:
                obj.setSelectedIndex([])

        self._dom_bdr_sel = (self._dom_bdr_sel[0],
                             tuple(i),
                             self._dom_bdr_sel[2],
                             self._dom_bdr_sel[3],)
        wx.CallAfter(self.canvas.refresh_hl)

    def highlight_edge(self, i, unselect=True):
        '''
        i is 1-based index
        '''
        try:
            x = len(i)
        except:
            i = list(i)
        ax = self.get_axes()

        if unselect:
            self.canvas.unselect_all()

        for name, obj in ax.get_children():
            if not name.startswith('edge'):
                continue
            if len(i) > 0:
                obj.setSelectedIndex(i)
                if len(obj._artists) > 0:
                    self.canvas.add_selection(obj._artists[0])
            else:
                obj.setSelectedIndex([])

        self._dom_bdr_sel = (self._dom_bdr_sel[0],
                             self._dom_bdr_sel[1],
                             tuple(i),
                             self._dom_bdr_sel[3],)

        wx.CallAfter(self.canvas.refresh_hl)

    def highlight_point(self, i, unselect=True):
        '''
        i is 1-based index
        '''
        try:
            x = len(i)
        except:
            i = list(i)
        ax = self.get_axes()

        if unselect:
            self.canvas.unselect_all()

        for name, obj in ax.get_children():
            if not name.startswith('point'):
                continue
            if len(i) > 0:
                obj.setSelectedIndex(i)
                if len(obj._artists) > 0:
                    self.canvas.add_selection(obj._artists[0])
            else:
                obj.setSelectedIndex([])

        self._dom_bdr_sel = (self._dom_bdr_sel[0],
                             self._dom_bdr_sel[1],
                             self._dom_bdr_sel[2],
                             tuple(i),)

        wx.CallAfter(self.canvas.refresh_hl)

    def highlight_none(self):
        self.canvas.unselect_all()

        ax = self.get_axes()
        for name, obj in ax.get_children():
            if hasattr(obj, 'setSelectedIndex'):
                obj.setSelectedIndex([])
        self.canvas.refresh_hl()

    def change_panel_button(self, kind):
        # kind = ('domain', 'face', 'edge', 'dot')
        if self._view_mode == 'geom':
            prefix = 'g'
        elif self._view_mode == 'mesh':
            prefix = 'm'
        else:
            prefix = ''
        self.canvas.toolbar.ClickP1Button(prefix+kind)

    def onResetModel(self, evt):
        ans = dialog.message(self,
                             "Do you want to delete all model setting?",
                             style=2)
        if ans == 'ok':
            self.model.scripts.helpers.reset_model()
            self.model.scripts.helpers.rebuild_ns()
            self.cla()
            if self.editdlg is not None:
                od = self.model.param.getvar('mfem_model')
                self.editdlg.set_model(od)

    def onEditModel(self, evt):
        from petram.pi.dlg_edit_model import DlgEditModel
        try:
            self.engine.assign_phys_pp_sel_index()
            self.engine.assign_sel_index()
        except:
            traceback.print_exc()

        model = self.model.param.getvar('mfem_model')
        if self.editdlg is None:
            self.model.scripts.helpers.rebuild_ns()
            self.editdlg = DlgEditModel(self, wx.ID_ANY, 'Model Tree',
                                        model=model)
            self.editdlg.Show()
        self.editdlg.Raise()

    def onSelectionPanel(self, evt):
        from petram.pi.selection_palette import SelectionPalette

        if self.selection_palette is None:
            self.selection_palette = SelectionPalette(
                self, wx.ID_ANY, 'Selection')
            self.selection_palette.Show()
        self.selection_palette.Raise()
        evt.Skip()

    def onGeomInfo(self, evt):
        try:
            from petram.geom.geom_info_palette import GeomInfoPalette

            if self.geom_info_palette is None:
                self.geom_info_palette = GeomInfoPalette(self,
                                                         wx.ID_ANY,
                                                         'Geometry')
                self.geom_info_palette.Show()
            self.geom_info_palette.Raise()
        except ImportError:
            pass
        evt.Skip()

    def onSaveModel(self, evt):
        from ifigure.widgets.dialog import write
        from petram.mesh.mesh_model import MeshFile
        path = write(parent=self,
                     message='Enter model file name',
                     wildcard='*.pmfm')
        if path == '':
            return
        self.model.scripts.helpers.save_model(path)

    def onSaveModelS(self, evt):
        from ifigure.widgets.dialog import writedir
        path = writedir(parent=self,
                        message='Directory to write')

        m = self.model.param.getvar('mfem_model')
        try:
            m.generate_script(dir=path)
        except:
            dialog.showtraceback(parent=self,
                                 txt='Failed to evauate expression',
                                 title='Error',
                                 traceback=traceback.format_exc())

    def onRunPreprocess(self, evt):
        try:
            self.run_preprocess()
        except:
            dialog.showtraceback(parent=self,
                                 txt='Failed to during pre-processing model data',
                                 title='Error',
                                 traceback=traceback.format_exc())

    def set_num_threads(self, nproc=1):
        import os
        import multiprocessing

        num_threads1 = self.model.param.getvar('openmp_num_threads')
        if num_threads1 is None:
            self.model.param.setvar('openmp_num_threads', 'auto')
            num_threads1 = 1
        if num_threads1 == 'auto':
            ncpus = multiprocessing.cpu_count()
            num_threads1 = ncpus//nproc

        os.environ["OMP_NUM_THREADS"] = str(num_threads1)

        num_threads2 = self.model.param.getvar('openblas_num_threads')
        if num_threads2 is None:
            self.model.param.setvar('openblas_num_threads', 1)
            num_threads2 = 1
        if num_threads2 != 'auto':
            os.environ["OPENBLAS_NUM_THREADS"] = str(num_threads2)

    def onSerDriver(self, evt):
        m = self.model.param.getvar('mfem_model')
        m.set_root_path(self.model.owndir())
        debug_level = m['General'].debug_level
        odir = os.getcwd()

        try:
            self.engine.run_verify_setting()
        except:
            os.chdir(odir)
            dialog.showtraceback(parent=self,
                                 txt='Failed to verify setting',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        try:
            self.run_preprocess()
        except:
            os.chdir(odir)
            dialog.showtraceback(parent=self,
                                 txt='Failed to during pre-processing model data',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        self.set_num_threads()
        self.model.scripts.run_serial.RunT(debug=debug_level)

        os.chdir(odir)

    def onParDriver(self, evt):
        m = self.model.param.getvar('mfem_model')
        m.set_root_path(self.model.owndir())
        debug_level = m['General'].debug_level
        odir = os.getcwd()

        try:
            self.engine.run_verify_setting()
        except:
            os.chdir(odir)
            dialog.showtraceback(parent=self,
                                 txt='Failed to verify setting',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        try:
            self.run_preprocess()
        except:
            os.chdir(odir)
            dialog.showtraceback(parent=self,
                                 txt='Failed to during pre-processing model data',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        nproc = self.model.param.getvar('nproc')
        if nproc is None:
            nproc = 2
        self.set_num_threads(nproc)

        self.model.scripts.run_parallel.RunT(nproc=nproc, debug=debug_level)

        os.chdir(odir)

    def viewer_canvasmenu(self):
        menus = [("+Petra-M", None, None), ]
        if self._hidemesh:
            menus.append(("Show Mesh",  self.onShowMesh, None))
        else:
            menus.append(("Hide Mesh",  self.onHideMesh, None))

        selmode = self.get_sel_mode()

        menus.append(("Copy " + selmode + " selection",
                      self.onCopySelection1, None))
        if self._view_mode == 'geom':
            menus.append(
                ("Copy " + selmode + " selection with prefix", self.onCopySelection2, None))

        # if len(self.canvas.selection) > 0:
        #    if self._view_mode == 'mesh':
        #        menus.append(("Show meshed " + selmode, self.onShowMeshedEntity, None))
        #        menus.append(("Hide meshed " + selmode, self.onHideMeshedEntity, None))

        if self.editdlg is not None and self._palette_focus == 'edit':
            check, kind, cidxs, labels = self.editdlg.isSelectionPanelOpen()
            if check:
                if kind == 'domain':
                    idx = self._dom_bdr_sel[0]
                elif kind == 'bdry':
                    idx = self._dom_bdr_sel[1]
                elif kind == 'edge':
                    idx = self._dom_bdr_sel[2]
                elif kind == 'point':
                    idx = self._dom_bdr_sel[3]
                else:
                    idx = None
                k = 0
                for cidx, label in zip(cidxs, labels):
                    if label == '':
                        continue
                    show_rm = any([x in cidx for x in idx])
                    show_add = any([not x in cidx for x in idx])

                    if show_add:
                        m = getattr(self, 'onAddSelection'+str(k))
                        txt = "Add to " + label
                        menus.append((txt, m, None))
                    if show_rm:
                        m = getattr(self, 'onRmSelection'+str(k))
                        txt = "Remove from " + label
                        menus.append((txt, m, None))
                    k = k + 1

        elif self.plotsoldlg is not None and self._palette_focus == 'plot':
            kind, cidx = self.plotsoldlg.get_selected_plotmode(kind=True)
            if kind == 'domain':
                idx = self._dom_bdr_sel[0]
            elif kind == 'bdry':
                idx = self._dom_bdr_sel[1]
            elif kind == 'edge':
                idx = self._dom_bdr_sel[2]
            elif kind == 'point':
                idx = self._dom_bdr_sel[3]
            else:
                idx = None
            if idx is not None:
                if cidx != 'all':
                    show_rm = any([x in cidx for x in idx])
                    show_add = any([not x in cidx for x in idx])
                else:
                    show_rm = False
                    show_add = False

                m1 = self.plotsoldlg.add_selection
                m2 = self.plotsoldlg.rm_selection
                m3 = self.plotsoldlg.set_selection

                def onAddPlotSel(evt, idx=idx):
                    m1(idx)
                    evt.Skip()

                def onRmPlotSel(evt, idx=idx):
                    m2(idx)
                    evt.Skip()

                def onSetPlotSel(evt, idx=idx):
                    m3(idx)
                    evt.Skip()
                if show_add:
                    menus.append(("Add to plot " + kind +
                                  " selection", onAddPlotSel, None))
                if show_rm:
                    menus.append(("Remove from plot " + kind + " selection", onRmPlotSel,
                                  None))
                menus.append(("Set to plot " + kind +
                              " selection", onSetPlotSel, None))
        elif self.selection_palette is not None and self._palette_focus == 'selection':
            pass
        else:
            pass

        palette_menu = []
        if self.selection_palette is None:
            palette_menu.append(
                ("Selection palette...", self.onSelectionPanel, None),)

        if (self._view_mode == 'geom' and
            self._view_mode_group.startswith('OCC') and
                self.geom_info_palette is None):
            palette_menu.append(("Geomtry info ...", self.onGeomInfo,  None),)

        if len(palette_menu) > 0:
            menus.append(('---', None, None),)
            menus.extend(palette_menu)

        menus.extend([("!", None, None),
                      ("---", None, None), ])
        return menus

    def _on_cp_selection1(self, use_prefix=False):

        kind = self.get_sel_mode()
        if kind == 'volume':
            idx = self._dom_bdr_sel[0]
            prefix = 'v'

        elif kind == 'face':
            idx = self._dom_bdr_sel[1]
            prefix = 'f'

        elif kind == 'edge':
            idx = self._dom_bdr_sel[2]
            prefix = 'l'

        elif kind == 'point':
            idx = self._dom_bdr_sel[3]
            prefix = 'p'

        else:
            pass

        prefix = prefix if use_prefix else ''

        return ", ".join([prefix + str(x) for x in idx])

    def onCopySelection1(self, evt):
        txt = self._on_cp_selection1(use_prefix=False)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(txt))
            wx.TheClipboard.Close()

    def onCopySelection2(self, evt):
        txt = self._on_cp_selection1(use_prefix=True)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(txt))
            wx.TheClipboard.Close()

    def onAddSelection(self, evt, flag=0):
        check, kind, cidx, labels = self.editdlg.isSelectionPanelOpen()
        if check:
            if kind == 'domain':
                idx = self._dom_bdr_sel[0]
            elif kind == 'bdry':
                idx = self._dom_bdr_sel[1]
            elif kind == 'edge':
                idx = self._dom_bdr_sel[2]
            elif kind == 'point':
                idx = self._dom_bdr_sel[3]
            else:
                idx = None
            if idx is not None:
                self.editdlg.add_remove_AreaSelection(idx, flag=flag)

    def onAddSelection0(self, evt):
        self.onAddSelection(evt, flag=0)

    def onAddSelection1(self, evt):
        self.onAddSelection(evt, flag=1)

    def onAddSelection2(self, evt):
        self.onAddSelection(evt, flag=2)

    def onRmSelection(self, evt, flag=0):
        check, kind, cidx, labels = self.editdlg.isSelectionPanelOpen()
        if check:
            if kind == 'domain':
                idx = self._dom_bdr_sel[0]
            elif kind == 'bdry':
                idx = self._dom_bdr_sel[1]
            elif kind == 'pair':
                idx = self._dom_bdr_sel[1]
            else:
                idx = None
            if idx is not None:
                self.editdlg.add_remove_AreaSelection(idx, rm=True, flag=flag)

    def onRmSelection0(self, evt):
        self.onRmSelection(evt, flag=0)

    def onRmSelection1(self, evt):
        self.onRmSelection(evt, flag=1)

    def onRmSelection2(self, evt):
        self.onRmSelection(evt, flag=2)

    def onSelVolume(self, evt):
        self.set_sel_mode('volume')

    def onSelFace(self, evt):
        self.set_sel_mode('face')

    def onSelEdge(self, evt):
        self.set_sel_mode('edge')

    def onSelPoint(self, evt):
        self.set_sel_mode('point')

    def onSelAny(self, evt):
        self.set_picker_mask('')
        self._sel_mode = ''

    def set_sel_mode(self, mode=None):
        if mode is None:
            self.refresh_toolbar_buttons()
            return
        self._sel_mode = mode
        mask = mode if mode != 'volume' else 'face'
        self.set_picker_mask(mask)

    def get_sel_mode(self):
        return self._sel_mode

    def set_picker_mask(self, key):
        for name, child in self.get_axes().get_children():
            child.set_pickmask(not name.startswith(key))

    def onShowMesh(self, evt=None):
        from petram.mesh.plot_mesh import plot_domainmesh
        mesh = self.engine.get_mesh()
        self._hidemesh = False
        self.update(False)

        children = [child
                    for name, child in self.get_axes().get_children()
                    if name.startswith('face')]
        childrene = [child
                     for name, child in self.get_axes().get_children()
                     if name.startswith('face') and name.endswith('e')]
        if len(childrene) > 0:
            for child in childrene:
                if not child.isempty():
                    child.set_linewidth(1.0, child._artists[0])
        else:
            for child in children:
                if not child.isempty():
                    child.set_linewidth(1.0, child._artists[0])
        self.update(True)
        self.draw_all()

    def onHideMesh(self, evt=None):
        self._hidemesh = True
        self.update(False)
        mesh = self.engine.get_mesh()
        children = [child
                    for name, child in self.get_axes().get_children()
                    if name.startswith('face')]
        for child in children:
            child.set_linewidth(0.0, child._artists[0])
        self.update(True)
        self.draw_all()

    def onShowAll(self, evt):
        for obj in self.book.walk_tree():
            if obj.is_suppress():
                obj.set_suppress(False)
        self.draw()

    def onPlotExpr(self, evt):
        try:
            self.plotexprdlg.Raise()
            return
        except AttributeError:
            pass
        except:
            traceback.print_exc()
            pass

        choices = self.model.param.getvar('mfem_model')['Phys'].keys()
        if len(choices) == 0:
            ans = dialog.message(self,
                                 "No physics is defined",
                                 style=0)
            return

        iattr = [x().figobj.name.split('_')[1]
                 for x in self.canvas.selection
                 if x().figobj.name.startswith('bdry')]

        from petram.pi.dlg_plot_expr import DlgPlotExpr
        # try:
        if self.plotexprdlg is not None:
            self.plotexprdlg.Raise()
        else:
            self.plotexprdlg = DlgPlotExpr(
                self, close_cb=self.onDlgPlotExprClose)

        evt.Skip()

    def onDlgPlotExprClose(self, evt):
        self.plotexprdlg = None
        evt.Skip()

    def onDlgPlotSol(self, evt):
        m = self.model.param.getvar('mfem_model')
        m.set_root_path(self.model.owndir())

        from petram.pi.dlg_plot_sol import DlgPlotSol
        # try:
        if self.plotsoldlg is not None:
            self.plotsoldlg.Raise()
        else:
            self.plotsoldlg = DlgPlotSol(self, close_cb=self.onDlgPlotSolClose)
        # except:
        #    self.plotsoldlg = DlgPlotSol(self)
        self.plotsoldlg.load_sol_if_needed()

    def onDlgPlotSolClose(self, evt):
        self.plotsoldlg = None
        evt.Skip()

    def onNewNS(self, evt):
        ret, txt = dialog.textentry(self,
                                    "Enter namespace name", "New NS...", '')
        if not ret:
            return
        self.model.scripts.helpers.create_ns(txt)
        from ifigure.events import SendChangedEvent
        SendChangedEvent(self.model, w=self)
        evt.Skip()

    def onLoadNS(self, evt):
        from petram.pi.ns_utils import import_ns
        import_ns(self.model)
        from ifigure.events import SendChangedEvent
        SendChangedEvent(self.model, w=self)
        evt.Skip()

    def onExportNS(self, evt):
        choices = ['_'.join(name.split('_')[:-1])
                   for name, child in self.model.namespaces.get_children()]
        if len(choices) == 0:
            return

        ll = [
            [None, choices[0], 4, {'style': wx.CB_READONLY,
                                   'choices': choices}], ]
        from ifigure.utils.edit_list import DialogEditList
        ret = DialogEditList(ll, modal=True, parent=self,
                             title='Select Namespace')
        if not ret[0]:
            return
        name = str(ret[1][0])

        from petram.pi.ns_utils import export_ns
        export_ns(self.model, name)
        from ifigure.events import SendChangedEvent
        SendChangedEvent(self.model, w=self)
        evt.Skip()

    def onRebuildNS(self, evt):
        try:
            self.rebuild_ns()
        except:
            dialog.showtraceback(parent=self,
                                 txt='Failed to rebuild namespace',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        dialog.message(parent=self,
                       message='Namespace is built successfully. ',
                       title='Passed',
                       center_on_screen=True)

        evt.Skip()

    def onClearSol(self, evt):
        self.model.scripts.helpers.clear_sol(w=self)
        self.model.param.setvar('sol', None)
        evt.Skip()

    def onNewLocalSol(self, evt):
        import re

        model = self.model
        if not model.has_child('solutions'):
            model.add_folder('solutions')

        param = model.param
        sol = param.eval('sol')
        if sol is None:
            names = sorted([x[0] for x in model.solutions.get_children()])
        else:
            names = [sol.name]

        if len(names) > 0:
            numbers = [''.join(re.findall('\d+$', n)) for n in names]
            numbers = [int(x if len(x) > 0 else '0') for x in numbers]
            basename = [n.rstrip('0123456789') for n in names]
            txt = basename[-1] + str(numbers[-1]+1)
        else:
            txt = 'sol'

        from ifigure.widgets.dialog import textentry
        f, txt = textentry(self, message='Enter local sol name',
                           title='Creating sol directory',
                           def_string=txt, center=True)
        if not f:
            return
        if not txt in names:
            folder = model.solutions.add_folder(txt)
            folder.mk_owndir()
        else:
            folder = model.solutions.get_child(name=txt)

        param.setvar('sol', '='+folder.get_full_path())

    # def onActivate(self, evt):
    #    windows = [self.editdlg, self.plotsoldlg, self.plotexprdlg]
    #    for w in windows:
    #        if w is not None:
    #            if not w.IsShown():
    #                w.Show()
    #    evt.Skip()

    def onWindowClose(self, evt=None):
        if self.editdlg is not None:
            try:
                self.editdlg.Destroy()
            except:
                pass
            self.editdlg = None
        if self.plotsoldlg is not None:
            try:
                self.plotsoldlg.Destroy()
            except:
                pass
            self.plotsoldlg = None

        super(MFEMViewer, self).onWindowClose(evt)

    def onServerSetting(self, evt):
        '''
        param needs to have two parameters
        remote = {'name': 'eofe7',
                  'rwdir': remote working directory,
                  'sol', None)
        host : points host setting object
        '''
        remote = self.model.param.eval('remote')
        if remote is not None:
            hostname = remote['name']
        else:
            hostname = ''

        proj = self.model.get_root_parent()
        p = proj.setting.parameters
        no_existing_c = True

        if p.hasvar('connection'):
            c = p.eval('connection')
            if c is not None:
                names = c.get_childnames()
                no_existing_c = (len(names) == 0)

        if no_existing_c:
            remote = {'name': '',
                      'rwdir': '',
                      'sol': ''}
            ret, new_name = dialog.textentry(self,
                                             "Enter the name of new connection",
                                             "Add Connection",
                                             hostname, center=False,
                                             center_on_screen=True)
        else:
            c = p.eval('connection')
            names = c.get_childnames()
            names.append("New...")
            ret, new_name = dialog.textselect(self,
                                              "Select the name of connection or New...",
                                              "Add Connection",
                                              def_string=hostname,
                                              center=False,
                                              center_on_screen=True,
                                              choices=names,
                                              endmodal_on_lastvalue=names[-1])
            if not ret:
                return
            if (ret and new_name == "New...") or c.get_child(name=new_name) is None:
                remote = {'name': '',
                          'rwdir': '',
                          'sol': ''}
                if new_name == "New...":
                    ret, new_name = dialog.textentry(self,
                                                     "Enter the name of new connection",
                                                     "Add Connection",
                                                     "enter-hostname.pppl.gov",
                                                     center=False,
                                                     center_on_screen=True)
                else:
                    pass
            else:
                child = c.get_child(name=new_name)
                child.onSetting()
                remote['name'] = new_name
                self.model.param.setvar('host', '='+child.get_full_path())
                return

        if ret:
            self.model.param.setvar('remote', remote)
            remote['name'] = new_name
            from petram.remote.client_script import make_remote_connection
            obj = make_remote_connection(self.model, new_name)
            self.model.param.setvar('host', '='+obj.get_full_path())

    def onServerSolve(self, evt):
        m = self.model.param.getvar('mfem_model')
        m.set_root_path(self.model.owndir())
        odir = os.getcwd()

        try:
            self.engine.run_verify_setting()
        except:
            os.chdir(odir)
            dialog.showtraceback(parent=self,
                                 txt='Failed to verify setting',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        try:
            self.run_preprocess()
        except:
            os.chdir(odir)
            dialog.showtraceback(parent=self,
                                 txt='Failed to during pre-processing model data',
                                 title='Error',
                                 traceback=traceback.format_exc())
            return

        remote = self.model.param.eval('remote')
        if remote is None:
            return

        from petram.pi.dlg_submit_job import get_defaults
        values, keys = get_defaults()

        for i, key in enumerate(keys):
            if remote.get(key, None) is not None:
                values[i] = remote.get(key, None)

        from petram.pi.dlg_submit_job import get_job_submisson_setting
        from petram.remote.client_script import get_job_queue

        dlg = progressbar(self, 'Checking queue config...',
                          'In progress', 5,
                          can_abort=True)
        dlg.Show()
        wx.GetApp().Yield()

        success, q = get_job_queue(self.model, progdlg=dlg)
        if dlg:
            dlg.Destroy()

        if not success:
            q = {'type': '',
                 'queues': [{'name': 'failed to read queue config'}, ]}

        if q["queues"][0]['name'] == ' no queue for a user':
            from ifigure.widgets.dialog import message
            ret = message(parent=self,
                          message="You do not have permission to access use to Petra-M on this computer system.",
                          title="No queue is available",
                          icon=wx.ICON_EXCLAMATION,
                          center_on_screen=False,
                          center_on_parent=True,)
            return

        setting = get_job_submisson_setting(self, remote['name'].upper(),
                                            value=values,
                                            queues=q)
        if len(setting) == 0:
            return

        dlg = progressbar(self, 'Preparing...', 'Job submission',
                          4, can_abort=True)
        dlg.Show()
        wx.GetApp().Yield()

        from petram.remote.client_script import prepare_remote_dir
        # if remote['rwdir'] != setting['rwdir']:
        cancelled = prepare_remote_dir(self.model,
                                       setting['rwdir'],
                                       dirbase='',
                                       progdlg=dlg)
        if not dlg:
            return
        if cancelled:
            dlg.Destroy()
            return

        for k in setting.keys():
            remote[k] = setting[k]
        if self.model.param.eval('sol') is None:
            sol = self.model.scripts.helpers.make_new_sol()
        else:
            sol = self.model.param.eval('sol')

        dlg.Update(1, newmsg="Preparing Data...")
        wx.GetApp().Yield()
        # remote.scripts.clean_remote_dir()
        sol.clean_owndir()
        self.model.scripts.helpers.save_model(os.path.join(sol.owndir(),
                                                           'model.pmfm'),
                                              meshfile_relativepath=True)

        from petram.remote.client_script import send_file, submit_job

        dlg.Update(2, newmsg="Sending file")
        wx.GetApp().Yield()

        cancelled = send_file(self.model, dlg,
                              skip_mesh=setting['skip_mesh'])
        if not dlg:
            return
        if cancelled:
            dlg.Destroy()
            return

        dlg.Update(3, newmsg="Submitting a job...")
        wx.GetApp().Yield()

        cancelled = submit_job(self.model, dlg)
        if not dlg:
            return
        if cancelled:
            dlg.Destroy()
            return

        dlg.Update(3, newmsg="Done...")
        wx.GetApp().Yield()
        wx.Sleep(1)

        dlg.Destroy()

    def onServerRetrieve(self, evt):
        from petram.remote.client_script import retrieve_files
        sol = self.model.param.eval('sol')
        if sol is None:
            sol = self.model.scripts.helpers.make_new_sol()
        sol_dir = sol.owndir()
        retrieve_files(self.model,
                       sol_dir=sol_dir)

    def run_preprocess(self):
        model = self.model
        engine = self.engine
        engine.preprocess_ns(model.namespaces, model.datasets)
        engine.build_ns()
        engine.run_preprocess(model.namespaces, model.datasets)

    def rebuild_ns(self):
        engine = self.engine
        model = self.model
        engine.preprocess_ns(model.namespaces, model.datasets)
        engine.build_ns()

    def get_internal_bc(self):
        d = self._s_v_loop['phys'][1]
        dd = []
        for k in d:
            dd.extend(d[k])

        from collections import defaultdict
        seen = defaultdict(int)
        for k in dd:
            seen[k] += 1
        return [k for k in seen if seen[k] > 1]

    def handle_shifthit(self):
        '''
        when space is hit in canvas
        '''
        if self._sel_mode == 'volume':
            mode = 'point'
        if self._sel_mode == 'face':
            mode = 'volume'
        if self._sel_mode == 'edge':
            mode = 'face'
        if self._sel_mode == 'point':
            mode = 'edge'
        self.set_toolbar_mode(mode)
        '''
        self._sel_mode = mode
        
        toolbarname = self.canvas.toolbar.p1_choice[1]
        buttonname = bmode + '_' + toolbarname

        if toolbarname == 'petram_mesh':
            buttonname = "m"+buttonname
        if toolbarname == 'petram_geom':
            buttonname = "g"+buttonname

        self.canvas.toolbar.ClickP1Button(buttonname)
        self.canvas.toolbar.ClickP1Button('select')
        '''

    def set_toolbar_mode(self, mode):
        bmodes = {'volume': 'domain',
                  'face': 'face',
                  'edge': 'edge',
                  'point': 'dot'}
        if not mode in bmodes:
            return
        bmode = bmodes[mode]

        self.set_sel_mode(mode)

        toolbarname = self.canvas.toolbar.p1_choice[1]
        buttonname = bmode + '_' + toolbarname

        if toolbarname == 'petram_mesh':
            buttonname = "m"+buttonname
        if toolbarname == 'petram_geom':
            buttonname = "g"+buttonname

        self.canvas.toolbar.ClickP1Button(buttonname)
        self.canvas.toolbar.ClickP1Button('select')



#
# this file is not used anymore...
#
from petram.helper.variables import var_g

g = var_g.copy()

def eval_value(value):
    return eval(value, g, {})    

def validator(value, param, ctrl):
    try:
       x = eval_value(value)
    except:
       return False
    return True



import os
import traceback
import gc

from petram.model import Model
from petram.solver.solver_model import Solver, SolveStep
from petram.namespace_mixin import NS_mixin
import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('Parametric')
format_memory_usage = debug.format_memory_usage

assembly_methods = {'Full assemble': 0,
                    'Reuse matrix': 1}

from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint
    from mpi4py import MPI
else:
    import mfem.ser as mfem
    nicePrint = dprint1

class Parametric(SolveStep, NS_mixin):
    '''
    parametric sweep of some model paramter
    and run solver
    '''
    can_delete = True
    has_2nd_panel = False

    def __init__(self, *args, **kwargs):
        SolveStep.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def init_solver(self):
        pass

    def panel1_param(self):
        v = self.get_panel1_value()
        return [["Initial value setting",   self.init_setting,  0, {}, ],
                ["Postprocess solution",    self.postprocess_sol,   0, {}, ],
                ["trial phys. ",   self.phys_model, 0, {}, ],
                ["assembly method",  'Full assemble',  4, {"readonly": True,
                                                           "choices": list(assembly_methods)}],
                self.make_param_panel('scanner',  v[2]),
                ["inner solver", '', 2, None],
                [None,  True,  3, {"text": "save separate mesh"}],
                [None, False, 3, {"text": "clear working dir."}],
                [None,  self.use_geom_gen,  3, {
                    "text": "run geometry generator"}],
                [None,  self.use_mesh_gen,  3, {"text": "run mesh generator"}],
                [None,  self.use_profiler,  3, {"text": "use profiler"}],
                ]

    def get_panel1_value(self):
        txt = list(assembly_methods)[0]
        for k, n in assembly_methods.items():
            if n == self.assembly_method:
                txt = k

        return (self.init_setting,
                self.postprocess_sol,
                self.phys_model,
                str(txt),
                str(self.scanner),
                self.get_inner_solver_names(),
                self.save_separate_mesh,
                self.clear_wdir,
                self.use_geom_gen,
                self.use_mesh_gen,
                self.use_profiler,)

    def import_panel1_value(self, v):
        self.init_setting = str(v[0])
        self.postprocess_sol = v[1]
        self.phys_model = str(v[2])
        self.assembly_method = assembly_methods[v[-8]]
        self.scanner = v[-7]
        self.save_separate_mesh = v[-5]
        self.clear_wdir = v[-4]
        self.use_geom_gen = v[-3]
        self.use_mesh_gen = v[-2]
        if self.use_geom_gen:
            self.use_mesh_gen = True
        if self.use_mesh_gen:
            self.assembly_method = 0
        self.use_profiler = bool(v[-1])

    def get_inner_solver_names(self):
        names = [s.name() for s in self.get_active_solvers()]
        return ', '.join(names)

    '''
    def get_inner_solvers(self):
        return [self[k] for k in self if self[k].enabled]
    '''

    def attribute_set(self, v):
        v = super(Parametric, self).attribute_set(v)
        v['assembly_method'] = 0
        v['scanner'] = 'Scan("a", [1,2,3])'
        v['save_separate_mesh'] = False
        v['clear_wdir'] = True

        return v

    def get_possible_child(self):
        #from solver.solinit_model import SolInit
        from petram.solver.std_solver_model import StdSolver
        from petram.solver.nl_solver_model import NLSolver
        from petram.solver.ml_solver_model import MultiLvlStationarySolver
        from petram.solver.solver_controls import DWCCall, ForLoop
        from petram.solver.set_var import SetVar

        try:
            from petram.solver.std_meshadapt_solver_model import StdMeshAdaptSolver
            return [MultiLvlStationarySolver,
                    StdSolver,
                    StdMeshAdaptSolver,
                    NLSolver,
                    DWCCall, ForLoop, SetVar]
        except:
            return [MultiLvlStationarySolver,
                    StdSolver,
                    NLSolver,
                    DWCCall, ForLoop, SetVar]

    def get_possible_child_menu(self):
        #from solver.solinit_model import SolInit
        from petram.solver.std_solver_model import StdSolver
        from petram.solver.nl_solver_model import NLSolver
        from petram.solver.ml_solver_model import MultiLvlStationarySolver
        from petram.solver.solver_controls import DWCCall, ForLoop
        from petram.solver.set_var import SetVar

        try:
            from petram.solver.std_meshadapt_solver_model import StdMeshAdaptSolver
            return [("", StdSolver),
                    ("", MultiLvlStationarySolver),
                    ("", NLSolver),
                    ("extra", ForLoop),
                    ("", StdMeshAdaptSolver),
                    ("", DWCCall),
                    ("!", SetVar)]
        except:
            return [("", StdSolver),
                    ("", MultiLvlStationarySolver),
                    ("", NLSolver),
                    ("extra", ForLoop),
                    ("", DWCCall),
                    ("!", SetVar)]

    def get_scanner(self, nosave=False):
        try:
            scanner = self.eval_param_expr(str(self.scanner),
                                           'scanner')[0]
            scanner.set_data_from_model(self.root())
        except:
            traceback.print_exc()
            return

        if not nosave:
            scanner.save_scanner_data(self)

        return scanner

    def get_default_ns(self):
        from petram.solver.parametric_scanner import Scan
        return {'Scan': Scan}

    def go_case_dir(self, engine, ksol, mkdir):
        '''
        make case directory and create symlinks
        '''

        od = os.getcwd()

        nsfiles = [n for n in os.listdir() if n.endswith('_ns.py')
                   or n.endswith('_ns.dat')]

        path = os.path.join(od, 'case_' + str(ksol))
        if mkdir:
            engine.mkdir(path)
            os.chdir(path)
            engine.cleancwd()
        else:
            os.chdir(path)
        files = ['model.pmfm'] + nsfiles
        for n in files:
            engine.symlink(os.path.join('../', n), n)
        self.case_dirs.append(path)
        return od

    def _run_full_assembly(self, engine, solvers, scanner, is_first=True):

        postprocess = self.get_pp_setting()
        for kcase, case in enumerate(scanner):
            is_first0 = True

            od = self.go_case_dir(engine, kcase, True)

            is_new_mesh = self.check_and_run_geom_mesh_gens(engine)

            if is_new_mesh or is_first:
                engine.preprocess_modeldata()
                is_first = False

            self.prepare_form_sol_variables(engine)

            self.init(engine)

            for ksolver, s in enumerate(solvers):
                is_first0 = s.run(engine, is_first=is_first0)
                engine.add_FESvariable_to_NS(self.get_phys())
                engine.store_x()
                if self.solve_error[0]:
                    dprint1("Parametric failed " + self.name() + ":" +
                            self.solve_error[1])

            engine.run_postprocess(postprocess, name=self.name())
            os.chdir(od)

    def _run_rhs_assembly(self, engine, solvers, scanner, is_first=True):
        self.prepare_form_sol_variables(engine)
#        self.init(engine, skip_essential=True)

        l_scan = len(scanner)

        all_phys = self.get_phys()
        phys_target = self.get_target_phys()
        postprocess = self.get_pp_setting()

        linearsolver = None
        for ksolver, s in enumerate(solvers):
            RHS_ALL = []
            instance = s.allocate_solver_instance(engine)

            phys_target = self.get_phys()
            phys_range = self.get_phys_range()

            for kcase, case in enumerate(scanner):

                if kcase == 0:
                    if ksolver == 0:
                        self.init(engine)
                    instance.set_blk_mask()
                    instance.assemble(inplace=False)
                else:
                    engine.set_update_flag('ParametricRHS')

                    done = []
                    for phys in phys_target + phys_range:
                        if phys not in done:
                            engine.run_update_param(phys)
                            done.append(phys)

                    engine.run_apply_essential(phys_target,
                                               phys_range,
                                               update=True)
                    engine.run_fill_X_block(update=True)
                    engine.run_assemble_extra_rhs(phys_target, phys_range,
                                                  update=True)
                    engine.run_assemble_b(phys_target, update=True)
                    engine.run_assemble_blocks(instance.compute_A,
                                               instance.compute_rhs,
                                               inplace=False,
                                               update=True)

                A, X, RHS, Ae, B, M, depvars = instance.blocks
                mask = instance.blk_mask
                depvars = [x for i, x in enumerate(depvars)
                           if mask[0][i]]
                if kcase == 0:
                    ls_type = instance.ls_type
                    phys_real = not s.is_complex()
                    AA = engine.finalize_matrix(A, mask, not phys_real,
                                                format=ls_type)

                RHS_ALL.append(RHS)

                if kcase == l_scan-1:
                    BB = engine.finalize_rhs(RHS_ALL, A, X[0], mask,
                                             not phys_real,
                                             format=ls_type)

                    if linearsolver is None:
                        linearsolver = instance.allocate_linearsolver(s.is_complex(),
                                                                      engine)
                    linearsolver.SetOperator(AA,
                                             dist=engine.is_matrix_distributed,
                                             name=depvars)

                    XX = None
                    solall = linearsolver.Mult(BB, x=XX, case_base=0)

                    is_sol_central = (True if not use_parallel else
                                      any(MPI.COMM_WORLD.allgather(solall is None)))

                    if not phys_real and s.assemble_real:
                        if not is_sol_central:
                            assert False, "this operation is not permitted"
                        oprt = linearsolver.oprt
                        solall = instance.linearsolver_model.real_to_complex(solall,
                                                                             oprt)

                    for ksol in range(l_scan):
                        instance.configure_probes('')
                        if ksol == 0:
                            instance.save_solution(mesh_only=True,
                                                   save_parmesh=s.save_parmesh)
                            save_mesh_linkdir = None

                        if is_sol_central:
                            A.reformat_central_mat(solall, ksol, X[0], mask)
                        else:
                            A.reformat_distributed_mat(solall, ksol, X[0], mask)

                        instance.sol = X[0]
                        for p in instance.probe:
                            p.append_sol(X[0])

                        od = self.go_case_dir(engine,
                                              ksol,
                                              ksolver == 0)

                        instance.save_solution(ksol=ksol,
                                               skip_mesh=False,
                                               mesh_only=False,
                                               save_parmesh=s.save_parmesh,
                                               save_mesh_linkdir=save_mesh_linkdir)

                        if save_mesh_linkdir is None:
                            save_mesh_linkdir = os.getcwd()

                        engine.sol = instance.sol
                        instance.save_probe()

                        engine.add_FESvariable_to_NS(self.get_phys())
                        engine.store_x()

                        engine.run_postprocess(postprocess, name=self.name())
                        os.chdir(od)

    def collect_probe_signals(self, dirs, scanner):
        from petram.sol.probe import list_probes, load_probe,  Probe
        params = scanner.list_data()

        od = os.getcwd()

        filenames, probenames = list_probes(dirs[0])

        names = scanner.names
        probes = [Probe(n, xnames=names) for n in probenames]

        for param, dirname in zip(params, dirs):
            os.chdir(dirname)
            for f, p in zip(filenames, probes):
                xdata, ydata = load_probe(f)
                p.append_value(ydata, param)

        os.chdir(od)
        for p in probes:
            p.write_file()
            # else:
            #    dprint1("skipping summarizing probe data for ", p.name)

    def set_scanner_physmodel(self, scanner):
        solvers = self.get_active_solvers()
        phys_models = []
        for s in solvers:
            for p in s.get_phys():
                if not p in phys_models:
                    phys_models.append(p)
        scanner.set_phys_models(phys_models)
        return solvers

    @debug.use_profiler
    def run(self, engine, is_first=True):
        #
        # is_first is not used
        #
        dprint1("Parametric Scan (assemly_methd=", self.assembly_method, ")")
        if self.clear_wdir:
            engine.remove_solfiles()

        engine.remove_case_dirs()

        scanner = self.get_scanner()
        if scanner is None:
            return

        solvers = self.set_scanner_physmodel(scanner)

        self.case_dirs = []
        if self.assembly_method == 0:
            self._run_full_assembly(
                engine, solvers, scanner, is_first=is_first)
        else:
            is_new_mesh = self.check_and_run_geom_mesh_gens(engine)
            if is_first or is_new_mesh:
                engine.preprocess_modeldata()
            self._run_rhs_assembly(engine, solvers, scanner, is_first=is_first)

        self.collect_probe_signals(self.case_dirs, scanner)
